from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class EMA(Indicator):
    """Exponential Moving Average - EMA

    The Exponential Moving Average is more responsive moving average compared to the
    Simple Moving Average (SMA).  The weights are determined by alpha which is
    proportional to it's length.

    Sources:
        https://www.investopedia.com/ask/answers/122314/what-exponential-moving-average-ema-formula-and-how-ema-calculated.asp

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 10
        source (str): Which input field to calculate the Indicator. Defaults to "close"
        smoothing (float): Smoothing multiplier for EMA. Defaults to 2.0
    """

    _name: str = field(init=False, default="EMA")
    period: int = 10
    source: str = "close"
    smoothing: float = 2.0
    _alpha: float = field(init=False, default=0)

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _validate_fields(self):
        self._alpha = float(self.smoothing / (self.period + 1.0))

    def _calculate_reading(self, index: int) -> float | dict | None:
        prev_ema = self.prev_reading()
        if prev_ema is not None:
            return float(
                self._alpha * self.reading(self.source) + (prev_ema * (1.0 - self._alpha))
            )

        if self.reading_period(self.period, self.source):
            return self.candles_average(self.period, self.source)

        return None



from typing import List, Optional

from hexital.analysis import utils
from hexital.core.candle import Candle
from hexital.utils.indexing import absindex


def doji(
    candles: List[Candle],
    lookback: Optional[int] = None,
    index: Optional[int] = None,
) -> bool:
    """Doji Pattern
    A candle body is Doji when it's shorter than 10% of the average of the
    n(10) previous candles' high-low range.

    Source: https://github.com/TA-Lib/ta-lib

    Args:
        candles (Optional[int]): Candles to use to find Doji Candle
        lookback (Optional[int]): Lookback allows detecting an Doji candles N back. Defaults to None.
        index (Optional[int]): Index of Candle to check. Defaults to None/Latest.

    Returns:
        bool: True if given Candle/Candle range is Doji
    """
    index_ = absindex(index, len(candles))

    if lookback is None:
        return _doji(candles, index_)

    return any(_doji(candles, i) for i in range(len(candles) - lookback, len(candles)))


def _doji(candles: List[Candle], index: int):
    if index < 10:
        return False
    return candles[index].realbody < utils.candle_doji(candles, index)


def dojistar(
    candles: List[Candle],
    lookback: Optional[int] = None,
    index: Optional[int] = None,
) -> bool:
    """Dojistar Pattern
    A Dojistar is either bearish or bullish, and is detected when we have a larger then average
    candle followed by a candle candle shorter than 10% of the average of the
    n(10) of the average and then the candle gaps up or down.
    A trend is required to find which direction.

    Source: https://github.com/TA-Lib/ta-lib
    Args:
        candles (Optional[int]): Candles to use to find Dojistar Candle
        lookback (Optional[int]): Lookback allows detecting an Dojistar candles N back. Defaults to None.
        index (Optional[int]): Index of Candle to check. Defaults to None/Latest.

    Returns:
        bool: True if given Candle/Candle range is Doji star
    """
    index_ = absindex(index, len(candles))

    if lookback is None:
        return _dojistar(candles, index_)

    return any(_dojistar(candles, i) for i in range(len(candles) - lookback, len(candles)))


def _dojistar(candles: List[Candle], index: int):
    if index < 10:
        return False
    candle = candles[index]
    prev_candle = candles[index - 1]

    if (
        prev_candle.realbody > utils.candle_bodylong(candles, index - 1)
        and candle.realbody <= utils.candle_doji(candles, index)
        and (
            (prev_candle.positive and utils.realbody_gapup(candle, prev_candle))
            or (prev_candle.negative and utils.realbody_gapdown(candle, prev_candle))
        )
    ):
        return True
    return False


def hammer(
    candles: List[Candle],
    lookback: Optional[int] = None,
    index: Optional[int] = None,
) -> bool | int:
    """Hammer Pattern
    A Hammer is detected when the Candle's open and close values are considered shorter
    than 10% of the average of the n(10) candles. However the low is larger than the average.

    Source: https://github.com/TA-Lib/ta-lib
    Args:
        candles (Optional[int]): Candles to use to find Hammer Candle
        lookback (Optional[int]): Lookback allows detecting an Hammer candles N back. Defaults to None.
        index (Optional[int]): Index of Candle to check. Defaults to None/Latest.

    Returns:
        bool: True if given Candle/Candle range is Hammer
    """
    index_ = absindex(index, len(candles))

    if lookback is None:
        return _hammer(candles, index_)

    return any(_hammer(candles, i) for i in range(len(candles) - lookback, len(candles)))


def _hammer(candles: List[Candle], index: int):
    if index < 10:
        return False
    candle = candles[index]

    if (
        candle.realbody < utils.candle_bodyshort(candles, index)
        and candle.shadow_lower > utils.candle_shadow_long(candles, index)
        and candle.shadow_upper < utils.candle_shadow_veryshort(candles, index)
        and min(candle.close, candle.open)
        <= candles[index - 1].low + utils.candle_near(candles, index - 1)
    ):
        return True

    return False


def inverted_hammer(
    candles: List[Candle],
    lookback: Optional[int] = None,
    index: Optional[int] = None,
) -> bool | int:
    """Inverted Hammer Pattern
    An Inverted Hammer is detected when the Candle's open and close values are considered shorter
    than 10% of the average of the n(10) candles. However the high is larger than the average.

    Source: https://github.com/TA-Lib/ta-lib
    Args:
        candles (Optional[int]): Candles to use to find Inverted Hammer Candle
        lookback (Optional[int]): Lookback allows detecting an InvertedHammer candles N back. Defaults to None.
        index (Optional[int]): Index of Candle to check. Defaults to None/Latest.

    Returns:
        bool: True if given Candle/Candle range is Inverted Hammer
    """
    index_ = absindex(index, len(candles))

    if lookback is None:
        return _inverted_hammer(candles, index_)

    return any(_inverted_hammer(candles, i) for i in range(len(candles) - lookback, len(candles)))


def _inverted_hammer(candles: List[Candle], index: int):
    if index < 10:
        return False
    candle = candles[index]
    prev_candle = candles[index - 1]

    if (
        candle.realbody < utils.candle_bodyshort(candles, index)
        and candle.shadow_upper > utils.candle_shadow_long(candles, index)
        and candle.shadow_lower < utils.candle_shadow_veryshort(candles, index)
        and utils.realbody_gapdown(candle, prev_candle)
    ):
        return True

    return False



from typing import List, Optional

from hexital.core.candle import Candle
from hexital.utils.candles import get_readings_period


def highest(
    candles: List[Candle], indicator: str, length: int, index: Optional[int] = None
) -> float | None:
    """
    Computes the highest value of the specified `indicator` over a given range of candles.
    The range includes the latest candle by default and considers up to the specified number of candles.
    """
    if index is None:
        index = len(candles) - 1

    readings = get_readings_period(candles, indicator, length, index, True)
    return max(readings, default=None)


def lowest(
    candles: List[Candle], indicator: str, length: int, index: Optional[int] = None
) -> float | None:
    """
    Computes the lowest value of the specified `indicator` over a given range of candles.
    The range includes the latest candle by default and considers up to the specified number of candles.
    """

    if not candles:
        return None
    if index is None:
        index = len(candles) - 1

    readings = get_readings_period(candles, indicator, length, index, True)
    return min(readings, default=None)


def realbody_avg(candles: List[Candle], length: int, index: Optional[int] = None) -> float:
    """
    Computes the average real body of a specified number of candles, including the current candle.
    The real body is calculated as the absolute difference between a candle's open and close prices.
    """

    if index is None:
        index = len(candles) - 1
    index += 1
    start_index = 0 if index - length < 0 else index - length

    return sum(candles[i].realbody for i in range(start_index, index)) / length


def high_low_avg(candles: List[Candle], length: int, index: Optional[int] = None) -> float:
    """
    Computes the average of the high-low range over a specified number of candles,
    including the current candle. The high-low range is the difference between a candle's
    high and low prices.
    """

    if index is None:
        index = len(candles) - 1
    index += 1
    start_index = 0 if index - length < 0 else index - length

    return sum(candles[i].high_low for i in range(start_index, index)) / length


def shadow_upper_avg(candles: List[Candle], length: int, index: Optional[int] = None) -> float:
    """
    Computes the average upper shadow over a specified number of candles, including the current candle.
    The upper shadow is the difference between a candle's high price and either its open or close price.
    """
    if index is None:
        index = len(candles) - 1
    index += 1
    start_index = 0 if index - length < 0 else index - length

    return sum(candles[i].shadow_upper for i in range(start_index, index)) / length


def shadow_lower_avg(candles: List[Candle], length: int, index: Optional[int] = None) -> float:
    """
    Computes the average lower shadow over a specified number of candles, including the current candle.
    The lower shadow is the difference between a candle's high price and either its open or close price.
    """
    if index is None:
        index = len(candles) - 1
    index += 1
    start_index = 0 if index - length < 0 else index - length

    return sum(candles[i].shadow_lower for i in range(start_index, index)) / length


def realbody_gapup(candle: Candle, candle_two: Candle) -> bool:
    """
    Computes if a candle has a real body gap-up compared to a previous candle.
    A gap-up occurs when the minimum value of the current candle's real body
    (i.e., the lower of its open or close) is greater than the maximum value
    of the previous candle's real body (i.e., the higher of its open or close).
    """

    return min(candle.open, candle.close) > max(candle_two.open, candle_two.close)


def realbody_gapdown(candle: Candle, candle_two: Candle) -> bool:
    """
    Computes if a candle has a real body gap-down compared to a previous candle.
    A gap-down occurs when the maximum value of the current candle's real body
    (i.e., the higher of its open or close) is less than the minimum value
    of the previous candle's real body (i.e., the lower of its open or close).
    """

    return max(candle.open, candle.close) < min(candle_two.open, candle_two.close)


def candle_gapup(candle: Candle, candle_two: Candle) -> bool:
    return candle.low > candle_two.high


def candle_gapdown(candle: Candle, candle_two: Candle) -> bool:
    return candle.high < candle_two.low


# Below are TA-lib globally used utils
# https://github.com/TA-Lib/ta-lib/blob/main/src/ta_common/ta_global.c


def _realbody_percentage(
    candles: List[Candle], index: Optional[int] = None, percentage: float = 1.0, length: int = 10
) -> float:
    if index is None:
        index = len(candles) - 1

    return realbody_avg(candles, length, index) * percentage


def _high_low_percentage(
    candles: List[Candle], index: Optional[int] = None, percentage: float = 1.0, length: int = 10
) -> float:
    if index is None:
        index = len(candles) - 1

    return high_low_avg(candles, length, index) * percentage


def candle_doji(candles: List[Candle], index: Optional[int] = None, length: int = 10) -> float:
    """real body is like doji's body when it's shorter than 10% the average of the 10 previous candles' high-low range"""
    return _high_low_percentage(candles, index=index, length=length, percentage=0.1)


def candle_bodylong(candles: List[Candle], index: Optional[int] = None, length: int = 10) -> float:
    """real body is long when it's longer than the average of the 10 previous candles' real body"""
    return _realbody_percentage(candles, index=index, length=length)


def candle_bodyverylong(
    candles: List[Candle], index: Optional[int] = None, length: int = 10
) -> float:
    """real body is very long when it's longer than 3 times the average of the 10 previous candles' real body"""
    return _realbody_percentage(candles, index=index, length=length, percentage=3)


def candle_bodyshort(
    candles: List[Candle], index: Optional[int] = None, length: int = 10
) -> float:
    """real body is short when it's shorter than the average of the 10 previous candles' real bodies"""
    return _realbody_percentage(candles, index=index, length=length)


def candle_shadow_veryshort(
    candles: List[Candle], index: Optional[int] = None, length: int = 10
) -> float:
    """shadow is very short when it's shorter than 10% the average of the 10 previous candles' high-low range"""
    return _high_low_percentage(candles, index=index, length=length, percentage=0.1)


def candle_shadow_short(
    candles: List[Candle], index: Optional[int] = None, length: int = 10
) -> float:
    """shadow is short when it's shorter than half the average of the 10 previous candles' sum of shadows"""
    return _high_low_percentage(candles, index=index, length=length)


def candle_shadow_long(candles: List[Candle], index: Optional[int] = None) -> float:
    """shadow is long when it's longer than the real body"""
    if index is None:
        index = -1

    return candles[index].realbody


def candle_shadow_verylong(candles: List[Candle], index: Optional[int] = None) -> float:
    """shadow is very long when it's longer than 2 times the real body"""
    if index is None:
        index = -1

    return candles[index].realbody * 2


def candle_near(candles: List[Candle], index: Optional[int] = None, length: int = 5) -> float:
    """when measuring distance between parts of candles or width of gaps
    near means "<= 20% of the average of the 5 previous candles' high-low range" """
    return _high_low_percentage(candles, index=index, length=length, percentage=0.2)


def candle_far(candles: List[Candle], index: Optional[int] = None, length: int = 5) -> float:
    """when measuring distance between parts of candles or width of gaps
    far means ">= 60% of the average of the 5 previous candles' high-low range"""
    return _high_low_percentage(candles, index=index, length=length, percentage=0.6)


def candle_equal(candles: List[Candle], index: Optional[int] = None, length: int = 5) -> float:
    """when measuring distance between parts of candles or width of gaps
    equal means "<= 5% of the average of the 5 previous candles' high-low range"""
    return _high_low_percentage(candles, index=index, length=length, percentage=0.05)



from abc import ABC, abstractmethod
from typing import List

from hexital.core.candle import Candle


class CandlestickType(ABC):
    name: str = "N/A"
    acronym: str = "NA"

    def __init__(self):
        return

    @abstractmethod
    def convert_candle(self, candle: Candle, candles: List[Candle], index: int): ...

    def conversion(self, candles: List[Candle]):
        for index in range(self._find_conv_index(candles), len(candles)):
            candle = candles[index]
            candle.save_clean_values()
            self.convert_candle(candle, candles, index)
            candle.reset_candle()
            candle.tag = self.name

    def _find_conv_index(self, candles: List[Candle]) -> int:
        if len(candles) == 0 or not candles[0].tag:
            return 0

        for index in range(len(candles) - 1, 0, -1):
            if self.name == candles[index].tag:
                return index + 1
        return len(candles)



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator, Managed
from hexital.indicators import EMA


@dataclass(kw_only=True)
class MACD(Indicator):
    """Moving Average Convergence Divergence - MACD

    The MACD is a popular indicator to that is used to identify a security's trend.
    While APO and MACD are the same calculation, MACD also returns two more series
    called Signal and Histogram. The Signal is an EMA of MACD and the Histogram is
    the difference of MACD and Signal.

    Sources:
        https://www.investopedia.com/ask/answers/122314/what-exponential-moving-average-ema-formula-and-how-ema-calculated.asp

    Output type: `Dict["MACD": float, "signal": float, "histogram": float]`

    Args:
        source (str): Which input field to calculate the Indicator. Defaults to "close"
        fast_period (int): How many Periods to use for fast EMA. Defaults to 12
        slow_period (int): How many Periods to use for slow EMA. Defaults to 26
        signal_period (int): How many Periods to use for MACD signal. Defaults to 9
    """

    _name: str = field(init=False, default="MACD")
    source: str = "close"
    fast_period: int = 12
    slow_period: int = 26
    signal_period: int = 9

    def _generate_name(self) -> str:
        return "{}_{}_{}_{}".format(
            self._name,
            self.fast_period,
            self.slow_period,
            self.signal_period,
        )

    def _validate_fields(self):
        if self.slow_period < self.fast_period:
            self.fast_period, self.slow_period = self.slow_period, self.fast_period

    def _initialise(self):
        self.data = self.add_managed_indicator("DATA", Managed(name=f"{self.name}_macd"))

        self.sub_emaf = self.add_sub_indicator(
            EMA(
                source=self.source,
                period=self.fast_period,
                name=f"{self.name}_EMA_fast",
            )
        )
        self.sub_emas = self.add_sub_indicator(
            EMA(
                source=self.source,
                period=self.slow_period,
                name=f"{self.name}_EMA_slow",
            )
        )
        self.sub_signal = self.add_managed_indicator(
            "signal",
            EMA(
                source=f"{self.name}_macd",
                period=self.signal_period,
                name=f"{self.name}_signal_line",
            ),
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        ema_slow = self.sub_emas.reading()

        if ema_slow is not None:
            macd = self.sub_emaf.reading() - ema_slow

            self.data.set_reading(macd)
            self.sub_signal.calculate_index(index)

            signal = self.sub_signal.reading()

            if signal is not None:
                histogram = macd - signal
                return {"MACD": macd, "signal": signal, "histogram": histogram}

        return {"MACD": None, "signal": None, "histogram": None}



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator
from hexital.indicators.stdev import STDEV


@dataclass(kw_only=True)
class STDEVT(Indicator):
    """Standard Deviation Threshold - STDEVT

    Standard Deviation while calculating threshold returning boolean signal
    if change to input is higher than threshold

    Sources:
        ChatGPT

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 10
        source (str): Which input field to calculate the Indicator. Defaults to "close"
        multiplier (float): A positive float to multiply the Deviation. Defaults to 2.0
    """

    _name: str = field(init=False, default="STDEVT")
    period: int = 10
    source: str = "close"
    multiplier: float = 2.0

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.sub_stdev = self.add_sub_indicator(
            STDEV(
                source=self.source,
                period=self.period,
                name=f"{self.name}_stdev",
            )
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        if not self.exists(f"{self.name}_stdev"):
            return False

        return (
            abs(self.reading(self.source) - self.prev_reading(self.source))
            > self.sub_stdev.reading() * self.multiplier
        )



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class Counter(Indicator):
    """Counter

    Simple Indictor which will count the current streak of a given value,
    specifically designed for bool values, but useable on any other re-occurring values.
    E.G Count the streak for current input value == count_value

    Output type: `float`

    Args:
        source (str): Which input field to calculate the Indicator.
        count_value (bool | int): Which value to be counting, E.G `bool`, `1`, etc
    """

    _name: str = field(init=False, default="COUNT")
    source: str
    count_value: bool | int = True

    def _generate_name(self) -> str:
        return f"{self._name}_{self.source.split('.')[0]}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        count = self.prev_reading(default=0)
        reading = self.reading(self.source, default=count)

        return count + 1 if self.count_value == reading else 0



from dataclasses import dataclass, field

from hexital.analysis import movement
from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class Donchian(Indicator):
    """Donchian Channels - Donchian

    Donchian Channels are a technical indicator that seeks to identify
    bullish and bearish extremes that favor reversals, higher and lower breakouts,
    breakdowns, and other emerging trends.

    Sources:
        https://upstox.com/learning-center/share-market/a-comprehensive-guide-to-donchian-channels-formula-calculation-and-strategic-uses/
        https://en.wikipedia.org/wiki/Donchian_channel

    Output type: `Dict["DCL": float, "DCM": float, "DCU": float]`

    Args:
        period (int): How many Periods to use. Defaults to 20
    """

    _name: str = field(init=False, default="DONCHIAN")
    period: int = 20

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        donchian = {"DCL": None, "DCM": None, "DCU": None}

        if self.prev_exists() or self.reading_period(self.period, "high", index):
            donchian["DCU"] = movement.highest(self.candles, "high", self.period - 1, index)
            donchian["DCL"] = movement.lowest(self.candles, "low", self.period - 1, index)
            donchian["DCM"] = (donchian["DCU"] + donchian["DCL"]) / 2

        return donchian



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator, Managed


@dataclass(kw_only=True)
class RSI(Indicator):
    """Relative Strength Index - RSI

    The Relative Strength Index is popular momentum oscillator used to measure the
    velocity as well as the magnitude of directional price movements.

    Sources:
        https://www.tradingview.com/support/solutions/43000502338-relative-strength-index-rsi/

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 14
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="RSI")
    period: int = 14
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.data = self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

    def _calculate_reading(self, index: int) -> float | dict | None:
        gains = None
        losses = None

        if self.prev_exists():
            change = self.prev_reading(self.source) - self.reading(self.source)

            change_gain = -1 * change if change < 0 else 0.0
            change_loss = change if change > 0 else 0.0

            gains = (
                (self.prev_reading(f"{self.name}_data.gain") * (self.period - 1)) + change_gain
            ) / self.period

            losses = (
                (self.prev_reading(f"{self.name}_data.loss") * (self.period - 1)) + change_loss
            ) / self.period
        elif self.reading_period(self.period + 1, self.source):
            changes = [
                self.reading(self.source, i) - self.reading(self.source, i - 1)
                for i in range(index - (self.period - 1), index + 1)
            ]

            gains = sum(chng for chng in changes if chng > 0) / self.period
            losses = sum(abs(chng) for chng in changes if chng < 0) / self.period

        self.data.set_reading({"gain": gains, "loss": losses})

        if gains is not None and losses is not None:
            return 100.0 - (100.0 / (1.0 + (gains / losses)))

        return None



from __future__ import annotations

from datetime import datetime, timedelta
from functools import cmp_to_key
from typing import List, Optional, Set

from hexital.core.candle import Candle
from hexital.core.candlestick_type import CandlestickType
from hexital.exceptions import InvalidCandleOrder
from hexital.utils.candles import reading_by_candle
from hexital.utils.timeframe import (
    clean_timestamp,
    on_timeframe,
    round_down_timestamp,
    timedelta_to_str,
)

DEFAULT_CANDLES = "default"


class CandleManager:
    _name: Optional[str] = None
    candles: List[Candle]
    candle_life: Optional[timedelta]
    timeframe: Optional[timedelta] = None
    timeframe_fill: bool = False
    candlestick: Optional[CandlestickType] = None

    def __init__(
        self,
        candles: Optional[List[Candle]] = None,
        candle_life: Optional[timedelta] = None,
        timeframe: Optional[timedelta] = None,
        timeframe_fill: bool = False,
        candlestick: Optional[CandlestickType] = None,
    ):
        if candles:
            self.candles = candles
        else:
            self.candles = []

        self.candle_life = candle_life
        self.timeframe = timeframe
        self.timeframe_fill = timeframe_fill
        self.candlestick = candlestick

        self._tasks()

    def __eq__(self, other) -> bool:
        if not isinstance(other, CandleManager):
            return False

        for key in ["candle_life", "timeframe", "timeframe_fill"]:
            if getattr(self, key) != getattr(other, key):
                return False

        return True

    @property
    def name(self) -> str:
        if self._name:
            return self._name
        elif self.timeframe:
            return timedelta_to_str(self.timeframe)
        else:
            return DEFAULT_CANDLES

    @name.setter
    def name(self, name: str):
        self._name = name

    def _tasks(self):
        self.collapse_candles()
        self.convert_candles()
        self.trim_candles()

    def find_indicator(self, name: str) -> bool:
        for candle in reversed(self.candles):
            if reading_by_candle(candle, name):
                return True
        return False

    def append(self, candles: Candle | List[Candle] | dict | List[dict] | list | List[list]):
        candles_ = []

        if isinstance(candles, Candle):
            candles_.append(candles)
        elif isinstance(candles, dict):
            candles_.append(Candle.from_dict(candles))
        elif isinstance(candles, list) and candles:
            candle_ = candles[0]
            if isinstance(candle_, Candle):
                candles_.extend(candles)
            elif isinstance(candle_, dict):
                candles_.extend(Candle.from_dicts(candles))
            elif isinstance(candle_, (float, int, datetime)):
                candles_.append(Candle.from_list(candles))
            elif isinstance(candle_, list):
                candles_.extend(Candle.from_lists(candles))
            else:
                raise TypeError

        self.sort_candles(candles_)

        to_sort = False
        last_timestamp = self.candles[-1].timestamp if self.candles else None

        for candle in candles_:
            if last_timestamp and candle.timestamp < last_timestamp:
                to_sort = True
            if not candle.timeframe or not self.timeframe or candle.timeframe <= self.timeframe:
                self.candles.append(candle.clean_copy())

        if to_sort:
            self.sort_candles()

        self._tasks()

    def sort_candles(self, candles: Optional[List[Candle]] = None):
        """Sorts Candles in order of timestamp, accounts for collapsing"""
        if candles:
            candles.sort(key=cmp_to_key(self._sort_comparison))
        else:
            self.candles.sort(key=cmp_to_key(self._sort_comparison))

    def _sort_comparison(self, candle_one: Candle, candle_two: Candle) -> int:
        """Sort's Candles in order but if timeframe exists, sorts with collapsing in mind.
        So mixed order of candles will arrange for candles to be collapsed down based on given timeframe
        EG:
            No Timeframe : [09:05:00 09:02:00 09:10:00] > [09:02:00 09:05:00 09:10:00]
            T5 Timeframe : [09:05:00 09:02:00 09:10:00] > [09:05:00 09:02:00 09:10:00]
        """
        time_one = candle_one.timestamp.timestamp()
        time_two = candle_two.timestamp.timestamp()

        if not self.timeframe:
            return int(time_one - time_two)

        timeframe = self.timeframe.total_seconds()

        if (
            time_two % timeframe
            and not time_one % timeframe
            and candle_two.timestamp > candle_one.timestamp - self.timeframe
        ) or (
            time_one % timeframe
            and not time_two % timeframe
            and candle_one.timestamp > candle_two.timestamp - self.timeframe
        ):
            return 1

        return int(time_one - time_two)

    def trim_candles(self):
        if self.candle_life is None or not self.candles:
            return

        latest = self.candles[-1].timestamp

        while self.candles[0].timestamp and self.candles[0].timestamp < latest - self.candle_life:
            self.candles.pop(0)

    def collapse_candles(self):
        """Collapses the given list of candles into specific timeframe candles.
        This can re-ran with same list to collapse latest candles.
        This method is destructive, returning a new list for the collapsed candles"""
        if not self.candles or not self.timeframe:
            return

        candles_ = [self.candles.pop(0)]
        init_candle = candles_[0]
        init_candle.timeframe = self.timeframe

        start_time = round_down_timestamp(init_candle.timestamp, self.timeframe)
        end_time = start_time + self.timeframe

        if not on_timeframe(init_candle.timestamp, self.timeframe):
            init_candle.set_collapsed_timestamp(end_time)

        while self.candles:
            candle = self.candles.pop(0)
            prev_candle = candles_[-1]

            next_end_time = end_time + self.timeframe
            candle.timestamp = clean_timestamp(candle.timestamp)
            candle.timeframe = self.timeframe

            if start_time < candle.timestamp <= end_time and prev_candle.timestamp == end_time:
                prev_candle.merge(candle)
            elif (
                start_time - self.timeframe < candle.timestamp <= start_time
                and prev_candle.timestamp == start_time
            ):
                prev_candle.merge(candle)
            elif start_time < candle.timestamp <= end_time:
                candle.set_collapsed_timestamp(end_time)
                candles_.append(candle)
            elif end_time < candle.timestamp <= next_end_time:
                candle.set_collapsed_timestamp(next_end_time)
                candles_.append(candle)
                start_time = end_time
                end_time = next_end_time
            elif start_time < candle.timestamp and on_timeframe(candle.timestamp, self.timeframe):
                start_time = round_down_timestamp(candle.timestamp, self.timeframe)
                end_time = start_time + self.timeframe
                candle.set_collapsed_timestamp(start_time)
                candles_.append(candle)
            elif next_end_time < candle.timestamp:
                start_time = round_down_timestamp(candle.timestamp, self.timeframe)
                end_time = start_time + self.timeframe
                candle.set_collapsed_timestamp(end_time)
                candles_.append(candle)
            else:
                # Shit's fucked yo
                raise InvalidCandleOrder(
                    f"Failed to collapse_candles due to invalid candle order prev: [{prev_candle}] - current: [{candle}]",
                )

        if self.timeframe_fill:
            candles_ = self._fill_timeframe_candles(candles_, self.timeframe)

        self.candles.extend(candles_)

    @staticmethod
    def _fill_timeframe_candles(candles: List[Candle], timeframe: timedelta) -> List[Candle]:
        if len(candles) < 2:
            return candles

        index = 1

        while index < len(candles):
            prev_candle = candles[index - 1]
            if candles[index].timestamp != prev_candle.timestamp + timeframe:
                fill_candle = Candle(
                    open=prev_candle.close,
                    close=prev_candle.close,
                    high=prev_candle.close,
                    low=prev_candle.close,
                    volume=0,
                    timestamp=prev_candle.timestamp + timeframe,
                    timeframe=prev_candle.timeframe,
                )
                fill_candle.aggregation_factor = 0
                candles.insert(index, fill_candle)

            index += 1

        return candles

    def convert_candles(self):
        if self.candles and self.candlestick:
            self.candlestick.conversion(self.candles)

    def purge(self, indicator: str | Set[str]):
        """Remove this indicator value from all Candles"""
        if isinstance(indicator, str):
            indicator = indicator

        for candle in self.candles:
            for name in indicator:
                candle.indicators.pop(name, None)
                candle.sub_indicators.pop(name, None)



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator, Managed
from hexital.indicators.hlca import HLCA


@dataclass(kw_only=True)
class MFI(Indicator):
    """Money Flow Index - MFI

    The money flow index (MFI) is an oscillator that ranges from 0 to 100.
    It is used to show the money flow over several days.

    Sources:
        https://www.tradingview.com/wiki/Money_Flow_(MFI)

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 14
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="MFI")
    period: int = 14
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.sub_hlca = self.add_sub_indicator(HLCA())
        self.data = self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

    def _calculate_reading(self, index: int) -> float | dict | None:
        hlca = self.sub_hlca.reading()
        prev_hlca = self.sub_hlca.prev_reading()

        money_flow = hlca * self.candles[index].volume

        if prev_hlca and hlca > prev_hlca:
            self.data.set_reading({"positive": money_flow, "negative": 0})
        elif prev_hlca and hlca < prev_hlca:
            self.data.set_reading({"positive": 0, "negative": money_flow})
        elif prev_hlca and hlca == prev_hlca:
            self.data.set_reading({"positive": 0, "negative": 0})

        if self.prev_exists() or self.sub_hlca.reading_period(self.period + 1, index=index):
            pos_money = self.candles_sum(self.period, f"{self.data.name}.positive")
            neg_money = self.candles_sum(self.period, f"{self.data.name}.negative")

            if pos_money and neg_money:
                return 100 * (pos_money / (pos_money + neg_money))

            return 0

        return None



from dataclasses import dataclass, field
from math import sqrt

from hexital.core.indicator import Indicator, Managed


@dataclass(kw_only=True)
class STDEV(Indicator):
    """Rolling Standard Deviation - STDEV

    You use a rolling stdev when you expect the standard deviation to change over time.
    As long as the standard deviation is changing slowly enough, we should be able to see
    the change in the standard deviation over time if we use the right size window.

    Sources:
        https://jonisalonen.com/2014/efficient-and-accurate-rolling-standard-deviation/

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 30
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="STDEV")
    period: int = 30
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

    def _calculate_reading(self, index: int) -> float | dict | None:
        popped_reading = 0

        reading = self.reading(self.source)

        if reading is None:
            return None

        if self.reading_period(self.period + 1, self.source, index):
            popped_reading = self.reading(self.source, index - self.period)

        old_mean = self.prev_reading(f"{self.name}_data.mean", 0.0)
        variance = self.prev_reading(f"{self.name}_data.variance", 0.0)

        mean_ = old_mean + (reading - popped_reading) / self.period

        variance += (
            (reading - popped_reading)
            * (reading - mean_ + popped_reading - old_mean)
            / (self.period)
        )

        self.managed_indicators["data"].set_reading({"mean": mean_, "variance": variance})

        if self.prev_exists() or self.reading_period(self.period, self.source, index):
            return sqrt(variance) if variance > 0 else 0



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class PivotPoints(Indicator):
    """Pivot Points - PP

    Pivot point is a price level that is used by traders as a possible indicator of market movement.
    A pivot point is calculated as an average of significant prices (high, low, close) from the
    performance of a market in the prior trading period. If the market in the following period trades
    above the pivot point it is 1usually evaluated as a bullish sentiment, whereas trading below
    the pivot point is seen as bearish.

    Sources:
        https://en.wikipedia.org/wiki/Pivot_point_(technical_analysis)

    Output type: `Dict["S1": float, "R1": float, "S2": float, "R2": float]`
    """

    _name: str = field(init=False, default="PP")

    def _generate_name(self) -> str:
        return f"{self._name}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        pivot_points = {"S1": None, "R1": None, "S2": None, "R2": None}

        if self.prev_exists("close"):
            high = self.candles[index - 1].high
            low = self.candles[index - 1].low
            close = self.candles[index - 1].close

            point = (high + low + close) / 3

            pivot_points["S1"] = (point * 2) - high
            pivot_points["R1"] = (point * 2) - low

            pivot_points["S2"] = point - (high - low)
            pivot_points["R2"] = point + (high - low)

        return pivot_points



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class VWMA(Indicator):
    """Volume Weighted Moving Average - VWMA

    VWMA is the ratio of the value of a security or financial asset traded to the total
    volume of transactions during a trading session. It is a measure of the average
    trading price for the period.

    Sources:
        https://www.investopedia.com/ask/answers/071414/whats-difference-between-moving-average-and-weighted-moving-average.asp

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 10
    """

    _name: str = field(init=False, default="VWMA")
    period: int = 10

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists() or self.reading_period(self.period, "close"):
            volume_close = sum(
                self.reading("close", i) * self.reading("volume", i)
                for i in range(index - (self.period - 1), index + 1)
            )
            return volume_close / self.candles_sum(self.period, "volume")
        return None



from __future__ import annotations

try:
    from datetime import UTC
except ImportError:
    UTC = False

from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

from hexital.exceptions import CandleAlreadyTagged
from hexital.utils.timeframe import TimeFrame, convert_timeframe_to_timedelta

IGNORE_CLEAN = ["indicators", "sub_indicators", "_clean_values", "_tag"]


class Candle:
    open: float
    high: float
    low: float
    close: float
    volume: int
    timestamp: datetime
    indicators: Dict[str, float | Dict[str, float | None] | None]
    sub_indicators: Dict[str, float | Dict[str, float | None] | None]
    timeframe: Optional[timedelta]
    aggregation_factor: int = 1
    _clean_values: Dict[str, float | int]
    _tag: Optional[str] = None
    _start_timestamp: Optional[datetime] = None
    _end_timestamp: Optional[datetime] = None

    def __init__(
        self,
        open: float,
        high: float,
        low: float,
        close: float,
        volume: int,
        timestamp: Optional[datetime | str] = None,
        timeframe: Optional[str | TimeFrame | timedelta | int] = None,
        indicators: Optional[Dict[str, float | Dict[str, float | None] | None]] = None,
        sub_indicators: Optional[Dict[str, float | Dict[str, float | None] | None]] = None,
    ):
        self.open = open
        self.high = high
        self.low = low
        self.close = close
        self.volume = volume
        self.timeframe = convert_timeframe_to_timedelta(timeframe) if timeframe else None

        if isinstance(timestamp, datetime):
            self.timestamp = timestamp
        elif isinstance(timestamp, str):
            self.timestamp = datetime.fromisoformat(timestamp)
        else:
            if UTC:
                self.timestamp = datetime.now(UTC)
            else:
                self.timestamp = datetime.utcnow()

        self._clean_values = {}
        self.indicators = indicators if indicators else {}
        self.sub_indicators = sub_indicators if sub_indicators else {}

    def __eq__(self, other) -> bool:
        if not isinstance(other, Candle):
            return False
        for key in set().union(self.__dict__.keys(), other.__dict__.keys()):
            if key in ["_start_timestamp", "_end_timestamp", "timeframe"]:
                local = getattr(self, key)
                remote = getattr(other, key)
                if remote is not None and local is not None and remote != local:
                    return False
            elif key.startswith("_"):
                continue
            elif getattr(self, key) != getattr(other, key):
                return False
        return True

    def __repr__(self) -> str:
        return str({name: value for name, value in vars(self).items() if not name.startswith("_")})

    @property
    def tag(self) -> str | None:
        return self._tag

    @tag.setter
    def tag(self, tag: str):
        if not self._tag:
            self._tag = tag
            return
        raise CandleAlreadyTagged(f"Candle already tagged as {self._tag} - [{self}]")

    @property
    def positive(self) -> bool:
        return self.open < self.close

    @property
    def negative(self) -> bool:
        return self.open > self.close

    @property
    def realbody(self) -> float:
        return abs(self.open - self.close)

    @property
    def shadow_upper(self) -> float:
        if self.positive:
            return abs(self.high - self.close)
        return abs(self.high - self.open)

    @property
    def shadow_lower(self) -> float:
        if self.positive:
            return abs(self.low - self.open)
        return abs(self.low - self.close)

    @property
    def high_low(self) -> float:
        return abs(self.high - self.low)

    def as_list(self) -> list:
        """
        Generates a list of values from the OHLCV values.
        `[timestamp, open, high, low, close, volume]` in that order.
        With an optional `timedelta` value at the end being the `timeframe`

        Returns:
            list: A list of the `Candle` values
        """
        return [self.timestamp, self.open, self.high, self.low, self.close, self.volume] + (
            [self.timeframe] if self.timeframe else []
        )

    def as_dict(self) -> dict:
        """
        Generates a dict of values from the OHLCV values, with  the following keys:
        `[timestamp, open, high, low, close, volume]`

        Returns:
            dict: A list of the `Candle` values
        """
        cdl = {
            "open": self.open,
            "high": self.high,
            "low": self.low,
            "close": self.close,
            "volume": self.volume,
            "timestamp": self.timestamp,
        }

        if self.timeframe:
            cdl["timeframe"] = self.timeframe

        return cdl

    @classmethod
    def from_dict(cls, candle: Dict[str, Any]) -> Candle:
        """
        Create a `Candle` object from a dictionary representation.

        The dictionary is expected to have the following keys:
        - Required: 'open', 'high', 'low', 'close', 'volume'
        - Optional: 'timestamp', 'time', 'date'
        - Optional: 'timeframe'

        The method extracts the values for these keys. If the optional keys for time ('timestamp',
        'time', etc.) are present, the first match is used as the timestamp.

        Args:
            candle (Dict[str, Any]): A dictionary containing the candle data.

        Returns:
            Candle: A `Candle` object initialized with the provided dictionary data.
        """
        time = [
            v
            for k, v in candle.items()
            if k in ["timestamp", "Timestamp", "time", "Time", "date", "Date"]
        ]
        return cls(
            candle.get("open", candle.get("Open", 0.0)),
            candle.get("high", candle.get("High", 0.0)),
            candle.get("low", candle.get("Low", 0.0)),
            candle.get("close", candle.get("Close", 0.0)),
            candle.get("volume", candle.get("Volume", 0)),
            timestamp=time[0] if time else None,
            timeframe=candle.get("timeframe", candle.get("Timeframe")),
        )

    @staticmethod
    def from_dicts(candles: List[Dict[str, float]]) -> List[Candle]:
        """
        Create's a list of `Candle` object's from a list of dictionary representation.

        Each dictionary is expected to have the following keys:
        - Required: 'open', 'high', 'low', 'close', 'volume'
        - Optional: 'timestamp', 'time', 'date'
        - Optional: 'timeframe'

        The method extracts the values for these keys. If the optional keys for time ('timestamp',
        'time', etc.) are present, the first match is used as the timestamp.
        Returning a list of `Candle` objects initialized with the provided dictionary data.

        Args:
            candles (List[Dict[str, Any]]): A dictionary containing the candle data.

        Returns:
            List[Candle]: A list of `Candle` object's.
        """
        return [Candle.from_dict(candle) for candle in candles]

    @classmethod
    def from_list(cls, candle: list) -> Candle:
        """
        Create a `Candle` object from a list representation.

        The list is expected to contain the following elements:
        - Required: `[open, high, low, close, volume]` in that order.
        - Optional: A `timestamp` at the beginning of the list.
        - Optional: A `timeframe` at the end of the list.

        If the first element is a `str` or `datetime`, it is treated as the `timestamp`.
        If the last element is a `str`, `int`, `TimeFrame`, or `timedelta`, it is treated as the `timeframe`.

        Args:
            candle (list): A list containing the candle data.

        Returns:
            Candle: A `Candle` object initialized with the data from the list.
        """
        timestamp = None
        timeframe = None

        if len(candle) > 5 and isinstance(candle[0], (str, datetime)):
            timestamp = candle.pop(0)
        if len(candle) > 5 and isinstance(candle[-1], (str, int, TimeFrame, timedelta)):
            timeframe = candle.pop(-1)

        return cls(
            open=candle[0],
            high=candle[1],
            low=candle[2],
            close=candle[3],
            volume=candle[4],
            timestamp=timestamp,
            timeframe=timeframe,
        )

    @staticmethod
    def from_lists(candles: List[list]) -> List[Candle]:
        """
        Create a list of `Candle` object's from a list of list representation.

        Each list is expected to contain the following elements:
        - Required: `[open, high, low, close, volume]` in that order.
        - Optional: A `timestamp` at the beginning of the list.
        - Optional: A `timeframe` at the end of the list.

        If the first element is a `str` or `datetime`, it is treated as the `timestamp`.
        If the last element is a `str`, `int`, `TimeFrame`, or `timedelta`, it is treated as the `timeframe`.

        Args:
            candles (List[list]): A list of list's containing the candle data.

        Returns:
            List[Candle]: A list of `Candle` object's.
        """
        return [Candle.from_list(candle) for candle in candles]

    def clean_copy(self) -> Candle:
        candle = Candle.from_list(self.as_list())
        candle.aggregation_factor = self.aggregation_factor
        return candle

    def set_collapsed_timestamp(self, timestamp: datetime):
        if not self._start_timestamp:
            self._start_timestamp = self.timestamp
        self.timestamp = timestamp

    def save_clean_values(self):
        self._clean_values = {
            name: value for name, value in vars(self).items() if name not in IGNORE_CLEAN
        }

    def recover_clean_values(self):
        if self._clean_values:
            for name, value in self._clean_values.items():
                self.__setattr__(name, value)
        self._clean_values = {}

    def reset_candle(self):
        self.indicators = {}
        self.sub_indicators = {}
        self._tag = None

    def merge(self, candle: Candle):
        """
        Merge another `Candle` object into the current candle.

        This method updates the current candle by integrating data from the provided `candle`.
        It ensures that the merged values respect the timeframe boundaries and adjusts
        attributes such as open, high, low, close, volume, and timestamps accordingly.

        **Note:**
        - Any calculated indicators will be wiped, as merging modifies the core candle values.
        - Any conversion or derived values associated with the candle will also be removed.

        Args:
            candle (Candle): The `Candle` object to merge into the current candle.

        Behaviour:
            - Adjusts the `open` if the merged candle's timestamp is earlier than the start timestamp.
            - Updates the `close` if the merged candle's timestamp is more recent.
            - Updates `high` and `low` based on the maximum and minimum values of the two candles.
            - Increases the `volume` by the volume of the merged candle.
            - Increments the `aggregation_factor` to account for the merged data.
            - Resets calculated indicators and cleans any derived values.
        """

        if self.timeframe:
            if (candle.timestamp + self.timeframe > self.timestamp + self.timeframe) or (
                candle.timestamp < self.timestamp - self.timeframe
            ):
                return

        self.recover_clean_values()

        if self._start_timestamp and candle.timestamp < self._start_timestamp:
            self.open = candle.open
            if not self._end_timestamp:
                self._end_timestamp = self._start_timestamp
            self._start_timestamp = candle.timestamp
        elif (
            self._start_timestamp
            and not self._end_timestamp
            and candle.timestamp > self._start_timestamp
        ):
            self.close = candle.close
            self._end_timestamp = candle.timestamp
        elif (
            self._start_timestamp
            and self._end_timestamp
            and candle.timestamp > self._end_timestamp
        ):
            self.close = candle.close
            self._end_timestamp = candle.timestamp
        elif (
            self._start_timestamp
            and self._end_timestamp
            and self._start_timestamp > candle.timestamp > self._end_timestamp
        ):
            pass
        else:
            self.close = candle.close

        self.high = max(self.high, candle.high)
        self.low = min(self.low, candle.low)
        self.volume += candle.volume
        self.aggregation_factor += candle.aggregation_factor

        self.reset_candle()



from dataclasses import dataclass, field

from hexital.analysis.utils import highest, lowest
from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class HL(Indicator):
    """Highest Lowest - HL

    Simple utility indicator to record and display the highest and lowest values N periods back.

    Output type: `Dict["low": float, "high": float]`

    Args:
        period (int): How many Periods to use. Defaults to 100
    """

    _name: str = field(init=False, default="HL")
    period: int = 100

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        return {
            "low": lowest(self.candles, "low", self.period, index),
            "high": highest(self.candles, "high", self.period, index),
        }



from dataclasses import dataclass, field

from hexital.analysis.utils import highest, lowest
from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class MOP(Indicator):
    """Midpoint Over Period - MOP

    A technical analysis tool that evaluates the average price movement by calculating
    the midpoint between the highest and lowest points over a specified period.
    This indicator aims to provide a smoother representation of price action,
    avoiding the choppiness of extreme highs and lows.

    Sources:
        https://trendspider.com/learning-center/understanding-and-applying-the-midpoint-over-period-indicator-in-trading/

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 2
        source (str()): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="MOP")
    period: int = 2
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists() or self.reading_period(self.period, self.source, index):
            return (
                lowest(self.candles, self.source, self.period, index)
                + highest(self.candles, self.source, self.period, index)
            ) * 0.5
        return None



from typing import Dict


def absindex(index: int | None, length: int) -> int:
    """Ensure's Index is a positive index, -1 == length-1"""
    if index is None:
        return length - 1
    if not valid_index(index, length):
        return length - 1
    if index < 0:
        return length + index
    return index


def valid_index(index: int | None, length: int) -> bool:
    if index is None:
        return False
    if not length > index >= -length:
        return False
    return True


def round_values(
    value: float | Dict[str, float | None] | None, round_by: int | None = 4
) -> float | Dict[str, float | None] | None:
    if round_by is None:
        return value

    if isinstance(value, float):
        return round(value, round_by)

    if isinstance(value, dict):
        for key, val in value.items():
            if isinstance(val, float):
                value[key] = round(val, round_by)

    return value



import math
from dataclasses import dataclass, field

from hexital.core.indicator import Indicator, Managed
from hexital.indicators.wma import WMA


@dataclass(kw_only=True)
class HMA(Indicator):
    """Hull Moving Average - HMA

    It is a combination of weighted moving averages designed
    to be more responsive to current price fluctuations while still smoothing prices.

    Sources:
        https://school.stockcharts.com/doku.php?id=technical_indicators:hull_moving_average

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 10
        source (str): Which input field to calculate the Indicator. Defaults to "close"

    """

    _name: str = field(init=False, default="HMA")
    period: int = 10
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.sub_wma = self.add_sub_indicator(
            WMA(
                source=self.source,
                period=self.period,
                name=f"{self.name}_WMA",
            )
        )
        self.sub_wmah = self.add_sub_indicator(
            WMA(
                source=self.source,
                period=int(self.period / 2),
                name=f"{self.name}_WMAh",
            )
        )

        self.sub_hma = self.add_managed_indicator("raw_HMA", Managed(name=f"{self.name}_HMAr"))
        self.sub_hma_smoothed = self.sub_hma.add_sub_indicator(
            WMA(
                source=f"{self.name}_HMAr",
                period=int(math.sqrt(self.period)),
                name=f"{self.name}_HMAs",
            ),
            False,
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        raw_hma = None
        wma = self.sub_wma.reading()

        if wma is not None:
            raw_hma = (2 * self.sub_wmah.reading()) - wma

        self.sub_hma.set_reading(raw_hma)
        return self.sub_hma_smoothed.reading()



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class WMA(Indicator):
    """Weighted Moving Average - WMA

    A Weighted Moving Average puts more weight on recent data and less on past data.
    This is done by multiplying each bar's price by a weighting factor.
    Because of its unique calculation, WMA will follow prices more closely
    than a corresponding Simple Moving Average.

    Sources:
        https://www.investopedia.com/ask/answers/071414/whats-difference-between-moving-average-and-weighted-moving-average.asp

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 10
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="WMA")
    period: int = 10
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists() or self.reading_period(self.period, self.source):
            values = sum(
                self.reading(self.source, i) * (self.period - py)
                for py, i in enumerate(range(index, index - self.period, -1))
            )
            weight = (self.period * (self.period + 1)) / 2
            return values / weight
        return None



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class ROC(Indicator):
    """Rate Of Change - ROC

    The Price Rate of Change (ROC) indicator in trading refers to the percentage change
    between the current price and the price of a set number of periods ago.
    It is used to identify the momentum of price movement and help traders make informed
    decisions regarding buying or selling assets. This indicator is calculated by dividing
    the difference between the current price and the price of a set number of periods ago
    by the previous price and multiplying by 100.

    Sources:
        https://en.wikipedia.org/wiki/Momentum_(technical_analysis)

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 10
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="ROC")
    period: int = 10
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists() or self.reading_period(self.period + 1, self.source):
            period_n_back = self.reading(self.source, index - self.period)

            if period_n_back == 0:
                return -100

            return ((self.reading(self.source) - period_n_back) / period_n_back) * 100
        return None



from dataclasses import dataclass, field

from hexital import indicators
from hexital.core.indicator import Indicator, Managed


@dataclass(kw_only=True)
class Supertrend(Indicator):
    """Supertrend

    It is used to identify market trends and potential entry and exit points in trading.
    The indicator is based on two dynamic values, period and multiplier, and incorporates
    the concept of Average True Range (ATR) to measure market volatility.
    The SuperTrend Indicator generates buy and sell signals by plotting a line on the price chart.

    Output type: `Dict["trend": float, "direction": int, "short": float]`

    Args:
        period (int): How many Periods to use. Defaults to 7
        source (str): Which input field to calculate the Indicator. Defaults to "close"
        multiplier (float): A positive float to multiply the ATR. Defaults to 3.0
    """

    _name: str = field(init=False, default="Supertrend")
    period: int = 7
    source: str = "close"
    multiplier: float = 3.0

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.sub_atr = self.add_sub_indicator(
            indicators.ATR(period=self.period, name=f"{self.name}_atr")
        )
        self.sub_hl = self.add_sub_indicator(indicators.HLA(name=f"{self.name}_HL"))
        self.data = self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

    def _calculate_reading(self, index: int) -> float | dict | None:
        direction = 1
        trend = None
        long = None
        short = None

        atr_ = self.sub_atr.reading()

        if atr_ is not None:
            mid_atr = self.multiplier * atr_

            upper = self.sub_hl.reading() + mid_atr
            lower = self.sub_hl.reading() - mid_atr

            prev_upper = self.prev_reading(f"{self.name}_data.upper")
            prev_lower = self.prev_reading(f"{self.name}_data.lower")

            if self.prev_exists(f"{self.name}_data.lower"):
                if self.candles[index].close > prev_upper:
                    direction = 1
                elif self.candles[index].close < prev_lower:
                    direction = -1
                else:
                    direction = self.prev_reading(f"{self.name}.direction")
                    if direction == 1 and lower < prev_lower:
                        lower = prev_lower
                    if direction == -1 and upper > prev_upper:
                        upper = prev_upper

            self.data.set_reading({"upper": upper, "lower": lower})

            trend = lower if direction == 1 else upper
            long = lower if direction == 1 else None
            short = upper if direction == -1 else None

        return {"trend": trend, "direction": direction, "long": long, "short": short}



from dataclasses import dataclass, field

from hexital.analysis import movement
from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class AROON(Indicator):
    """Aroon - AROON

    The Aroon indicator, indicates if a price is trending or is in a trading range.
    It can also reveal the beginning of a new trend, its strength and can help anticipate
    changes from trading ranges to trends.

    Sources:
        https://www.fidelity.com/learning-center/trading-investing/technical-analysis/technical-indicator-guide/aroon-indicator


    Output type: `Dict["AROONU": float, "AROOND": float, "AROONOSC": float]`

    Args:
        period (int): How many Periods to use. Defaults to 14
    """

    _name: str = field(init=False, default="AROON")
    period: int = 14

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        aroon = {
            "AROONU": None,
            "AROOND": None,
            "AROONOSC": None,
        }
        if self.prev_exists() or self.reading_period(self.period + 1, "high"):
            aroon["AROONU"] = (
                (self.period - movement.highestbar(self.candles, "high", self.period + 1, index))
                / self.period
            ) * 100
            aroon["AROOND"] = (
                (self.period - movement.lowestbar(self.candles, "low", self.period + 1, index))
                / self.period
            ) * 100

            aroon["AROONOSC"] = aroon["AROONU"] - aroon["AROOND"]

        return aroon



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator
from hexital.indicators.sma import SMA
from hexital.indicators.stdev import STDEV


@dataclass(kw_only=True)
class BBANDS(Indicator):
    """Bollinger Bands - BBANDS

    Bollinger Bands are a type of statistical chart characterizing
    the prices and volatility over time of a financial instrument or commodity,
    using a formulaic method.

    Sources:
        https://www.britannica.com/money/bollinger-bands-indicator

    Output type: `Dict["BBL": float, "BBM": float, "BBU": float]`

    Args:
        period (int): How many Periods to use. Defaults to 5
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="BBANDS")
    period: int = 5
    source: str = "close"
    _std: float = field(init=False, default=2.0)

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.sub_stdev = self.add_sub_indicator(STDEV(source=self.source, period=self.period))
        self.sub_sma = self.add_sub_indicator(SMA(source=self.source, period=self.period))

    def _calculate_reading(self, index: int) -> float | dict | None:
        bbands = {
            "BBL": None,
            "BBM": None,
            "BBU": None,
        }
        if self.prev_exists() or (self.sub_sma.exists() and self.sub_stdev.exists()):
            sma = self.sub_sma.reading()
            stdev = self.sub_stdev.reading()

            bbands["BBM"] = sma
            bbands["BBL"] = sma - (stdev * self._std)
            bbands["BBU"] = sma + (stdev * self._std)

        return bbands



from .adx import ADX
from .amorph import Amorph
from .aroon import AROON
from .atr import ATR
from .bbands import BBANDS
from .cmo import CMO
from .counter import Counter
from .donchian import Donchian
from .ema import EMA
from .highest_lowest import HL
from .hla import HLA
from .hlca import HLCA
from .hma import HMA
from .jma import JMA
from .kc import KC
from .macd import MACD
from .mfi import MFI
from .midpoint import MOP
from .obv import OBV
from .pivot_points import PivotPoints
from .rma import RMA
from .roc import ROC
from .rsi import RSI
from .rvi import RVI
from .sma import SMA
from .stdev import STDEV
from .stdevthres import STDEVT
from .stoch import STOCH
from .supertrend import Supertrend
from .tr import TR
from .tsi import TSI
from .vwap import VWAP
from .vwma import VWMA
from .wma import WMA



from typing import List, Optional

from hexital.analysis import utils
from hexital.core.candle import Candle
from hexital.core.hexital import Hexital
from hexital.core.indicator import Indicator
from hexital.utils.candles import get_readings_period, reading_by_candle, reading_by_index
from hexital.utils.indexing import absindex, valid_index
from hexital.utils.timeframe import within_timeframe


def _retrieve_candles(
    obj: Indicator | Hexital | List[Candle],
    indicator: Optional[str] = None,
    indicator_cmp: Optional[str] = None,
) -> List[List[Candle]]:
    if isinstance(obj, list):
        return [obj]
    elif isinstance(obj, Indicator):
        return [obj.candles]
    elif isinstance(obj, Hexital) and not indicator and not indicator_cmp:
        return [obj.candles("default")]
    elif isinstance(obj, Hexital) and indicator:
        return obj.find_candles(indicator, indicator_cmp)

    return []


def _timeframe_pair_candles(
    candles: List[List[Candle]],
) -> List[List[Candle]]:
    organised = [[], []]
    set_one = candles[0]
    set_two = candles[1]

    if set_one[-1].timeframe == set_two[-1].timeframe:
        return candles
    else:
        second_pointer = len(candles[1]) - 1

        for first_pointer in range(len(set_one) - 1, -1, -1):
            if second_pointer == 0:
                break

            while (
                not within_timeframe(
                    set_one[first_pointer].timestamp,
                    set_two[second_pointer].timestamp,
                    set_two[second_pointer].timeframe,
                )
                and not within_timeframe(
                    set_two[second_pointer].timestamp,
                    set_one[first_pointer].timestamp,
                    set_one[first_pointer].timeframe,
                )
            ) and second_pointer >= 0:
                second_pointer -= 1

            organised[0].insert(0, set_one[first_pointer])
            organised[1].insert(0, set_two[second_pointer])

    return organised


def positive(candles: Candle | List[Candle], index: int = -1) -> bool:
    if isinstance(candles, Candle):
        return candles.positive

    if not valid_index(index, len(candles)):
        return False
    return candles[index].positive


def negative(candles: Candle | List[Candle], index: int = -1) -> bool:
    if isinstance(candles, Candle):
        return candles.negative

    if not valid_index(index, len(candles)):
        return False
    return candles[index].negative


def above(
    candles: Indicator | Hexital | List[Candle],
    indicator: str,
    indicator_cmp: str,
    length: int = 0,
    index: int = -1,
) -> bool:
    """Above Analysis

    Checks whether the `indicator` reading is higher than the `indicator_cmp` reading.
    By default, it evaluates the latest candle but can also check `n` candles back.
    If any candle within the specified range is above, it returns `True`.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The primary indicator to evaluate.
        indicator_cmp (str): The secondary indicator to compare against.
        length (int, optional):  The number of candles to include in the range. Defaults to 0 (only the current index).
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if `indicator` is above `indicator_cmp` within the specified range; otherwise `False`.
    """
    candles_ = _retrieve_candles(candles, indicator, indicator_cmp)

    if not candles_:
        return False

    if len(candles_) == 1 and candles_[0]:
        candle_set = candles_[0]
        idx = absindex(index, len(candle_set)) + 1
        length = idx - (length + 1)

        return _above(
            candle_set[length:idx],
            indicator,
            candle_set[length:idx],
            indicator_cmp,
        )
    elif len(candles_) == 2:
        candle_set = _timeframe_pair_candles(candles_)
        idx = absindex(index, len(candle_set[0])) + 1
        length = idx - (length + 1)

        return _above(
            candle_set[0][length:idx],
            indicator,
            candle_set[1][length:idx],
            indicator_cmp,
        )

    return False


def _above(
    candles: List[Candle],
    indicator: str,
    candles_two: List[Candle],
    indicator_cmp: str,
) -> bool:
    if len(candles) != len(candles_two):
        return False

    for i in range(len(candles)):
        reading_one = reading_by_index(candles, indicator, i)
        reading_two = reading_by_index(candles_two, indicator_cmp, i)

        if isinstance(reading_one, (float, int)) and isinstance(reading_two, (float, int)):
            if reading_one > reading_two:
                return True

    return False


def below(
    candles: Indicator | Hexital | List[Candle],
    indicator: str,
    indicator_cmp: str,
    length: int = 0,
    index: int = -1,
) -> bool:
    """Below Analysis

    Checks whether the `indicator` reading is lower than the `indicator_cmp` reading.
    By default, it evaluates the latest candle but can also check `n` candles back.
    If any candle within the specified range is below, it returns `True`.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The primary indicator to evaluate.
        indicator_cmp (str): The secondary indicator to compare against.
        length (int, optional): The number of candles to include in the range. Defaults to 0 (only the current index).
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if `indicator` is below `indicator_cmp` within the specified range; otherwise `False`.
    """
    candles_ = _retrieve_candles(candles, indicator, indicator_cmp)

    if not candles_:
        return False

    if len(candles_) == 1 and candles_[0]:
        candle_set = candles_[0]
        idx = absindex(index, len(candle_set)) + 1
        length = idx - (length + 1)

        return _below(
            candle_set[length:idx],
            indicator,
            candle_set[length:idx],
            indicator_cmp,
        )
    elif len(candles_) == 2:
        candle_set = _timeframe_pair_candles(candles_)
        idx = absindex(index, len(candle_set[0])) + 1
        length = idx - (length + 1)

        return _below(
            candle_set[0][length:idx],
            indicator,
            candle_set[1][length:idx],
            indicator_cmp,
        )

    return False


def _below(
    candles: List[Candle],
    indicator: str,
    candles_two: List[Candle],
    indicator_cmp: str,
) -> bool:
    if len(candles) != len(candles_two):
        return False

    for i in range(len(candles)):
        reading_one = reading_by_index(candles, indicator, i)
        reading_two = reading_by_index(candles_two, indicator_cmp, i)

        if isinstance(reading_one, (float, int)) and isinstance(reading_two, (float, int)):
            if reading_one < reading_two:
                return True

    return False


def value_range(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 4, index: int = -1
) -> float | None:
    """Value Range Analysis

    Calculates the difference between the minimum and maximum values for the given `indicator`
    within a specified range of candles. Includes the latest candle by default. If the specified
    length exceeds the available candles, it will evaluate all candles.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 4.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        float | None: The difference between the minimum and maximum indicator values in the range,
        or `None` if there are insufficient readings.
    """
    candle_set = _retrieve_candles(candles, indicator)[0]
    if not candle_set:
        return None

    readings = get_readings_period(candle_set, indicator, length, index, True)

    if len(readings) < 2:
        return None

    return abs(min(readings) - max(readings))


def rising(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 1, index: int = -1
) -> bool:
    """Rising Analysis

    Determines whether the `indicator` consistently rises across a specified range of candles.
    By default, it checks if the current indicator value is greater than the previous one.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 1.
            (compares the latest with the previous).
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if the `indicator` is greater than each previous readings in the range; otherwise `False`.
    """
    candle_set = _retrieve_candles(candles, indicator)[0]

    idx = absindex(index, len(candle_set))

    if length < 1 or len(candle_set) < 2:
        return False

    latest_reading = reading_by_candle(candle_set[idx], indicator)
    if latest_reading is None or isinstance(latest_reading, dict):
        return False

    readings = get_readings_period(candle_set, indicator, length, idx)
    if not readings:
        return False

    for reading in readings:
        if reading >= latest_reading:
            return False
    return True


def falling(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 1, index: int = -1
) -> bool:
    """Falling Analysis

    Determines whether the `indicator` consistently falling across a specified range of candles.
    By default, it checks if the current indicator value is lower than the previous one.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 1.
            (compares the latest with the previous).
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if the `indicator` is lower than each previous readings in the range; otherwise `False`.
    """
    candle_set = _retrieve_candles(candles, indicator)[0]
    idx = absindex(index, len(candle_set))

    if length < 1 or len(candle_set) < 2:
        return False

    latest_reading = reading_by_candle(candle_set[idx], indicator)
    if latest_reading is None or isinstance(latest_reading, dict):
        return False

    readings = get_readings_period(candle_set, indicator, length, idx)
    if not readings:
        return False

    for reading in readings:
        if reading <= latest_reading:
            return False
    return True


def mean_rising(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 4, index: int = -1
) -> bool:
    """Mean Rising Analysis

    Evaluates whether the `indicator` is, on average, rising across a specified range of candles.
    By default, it checks if the current indicator value is higher than the average of the previous four readings.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 4.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if the `indicator` is higher than the average of the specified `n` readings; otherwise `False`.
    """

    candle_set = _retrieve_candles(candles, indicator)[0]
    idx = absindex(index, len(candle_set))

    if length < 1 or len(candle_set) < 2:
        return False

    latest_reading = reading_by_candle(candle_set[idx], indicator)
    if latest_reading is None or isinstance(latest_reading, dict):
        return False

    readings = get_readings_period(candle_set, indicator, length, idx)
    if not readings:
        return False

    return sum(readings) / len(readings) < latest_reading


def mean_falling(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 4, index: int = -1
) -> bool:
    """Mean Falling Analysis

    Evaluates whether the `indicator` is, on average, falling across a specified range of candles.
    By default, it checks if the current indicator value is lower than the average of the previous four readings.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 4.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if the `indicator` is lower than the average of the specified `n` readings; otherwise `False`.
    """
    candle_set = _retrieve_candles(candles, indicator)[0]
    idx = absindex(index, len(candle_set))

    if length < 1 or len(candle_set) < 2:
        return False

    latest_reading = reading_by_candle(candle_set[idx], indicator)
    if latest_reading is None or isinstance(latest_reading, dict):
        return False

    readings = get_readings_period(candle_set, indicator, length, idx)
    if not readings:
        return False

    return sum(readings) / len(readings) > latest_reading


def highest(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 4, index: int = -1
) -> float | None:
    """Highest Reading Analysis

    Determines the highest value of the specified `indicator` over a given number of candles.
    By default, includes the latest candle and evaluates up to the previous four candles.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 4.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        float | None: The highest reading for the specified `indicator` within the range,
        or `None` if no valid readings are found.
    """
    return utils.highest(_retrieve_candles(candles, indicator)[0], indicator, length, index)


def lowest(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 4, index: int = -1
) -> float | None:
    """Lowest Reading Analysis

    Determines the lowest value of the specified `indicator` over a given number of candles.
    By default, includes the latest candle and evaluates up to the previous four candles.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 4.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        float | None: The lowest reading for the specified `indicator` within the range,
        or `None` if no valid readings are found.
    """
    return utils.lowest(_retrieve_candles(candles, indicator)[0], indicator, length, index)


def highestbar(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 4, index: int = -1
) -> int | None:
    """Highest Bar Offset Analysis

    Determines the offset (distance) to the candle with the highest reading of the specified `indicator`
    within a given range. By default, includes the latest candle and evaluates up to the previous four candles.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 4.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        int | None: The offset to the candle with the highest reading, relative to the starting index,
        or `None` if no valid readings are found.
    """
    candle_set = _retrieve_candles(candles, indicator)[0]

    if not candles:
        return None

    idx = absindex(index, len(candle_set))

    high = None
    distance = 0

    for idx, index in enumerate(range(idx, idx - length, -1)):
        current = reading_by_index(candle_set, indicator, index)
        if not isinstance(current, (float, int)):
            continue

        if high is None:
            high = current

        if high < current:
            high = current
            distance = idx

    return distance


def lowestbar(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 4, index: int = -1
) -> int | None:
    """Lowest Bar Offset Analysis

    Determines the offset (distance) to the candle with the lowest reading of the specified `indicator`
    within a given range. By default, includes the latest candle and evaluates up to the previous four candles.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The name of the indicator to evaluate.
        length (int, optional): The number of candles to include in the range. Defaults to 4.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        int | None: The offset to the candle with the lowest reading, relative to the starting index,
        or `None` if no valid readings are found.
    """
    candle_set = _retrieve_candles(candles, indicator)[0]
    if not candle_set:
        return None

    idx = absindex(index, len(candle_set))

    low = None
    distance = 0

    for idx, index in enumerate(range(idx, idx - length, -1)):
        current = reading_by_index(candle_set, indicator, index)
        if not isinstance(current, (float, int)):
            continue

        if low is None:
            low = current

        if low > current:
            low = current
            distance = idx

    return distance


def cross(
    candles: Indicator | Hexital | List[Candle],
    indicator: str,
    indicator_cmp: str,
    length: int = 1,
    index: int = -1,
) -> bool:
    """Cross Analysis

    Determines whether the `indicator` reading has crossed the `indicator_cmp` reading
    within a specified range of candles. The cross can occur in either direction.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The primary indicator to evaluate.
        indicator_cmp (str): The secondary indicator to compare against.
        length (int, optional): The number of candles to include in the range. Defaults to 1.
            (compares the latest with the previous).
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if `indicator` has crossed `indicator_cmp` within the specified range; otherwise `False`.
    """
    candles_ = _retrieve_candles(candles, indicator, indicator_cmp)

    if not candles_:
        return False

    if len(candles_) == 1 and candles_[0]:
        candle_set = candles_[0]
        idx = absindex(index, len(candle_set)) + 1
        length = idx - (length + 1)

        return _cross(candle_set[length:idx], indicator, candle_set[length:idx], indicator_cmp)
    elif len(candles_) == 2:
        candle_set = _timeframe_pair_candles(candles_)
        idx = absindex(index, len(candle_set[0])) + 1
        length = idx - (length + 1)

        return _cross(
            candle_set[0][length:idx], indicator, candle_set[1][length:idx], indicator_cmp
        )

    return False


def _cross(
    candles: List[Candle],
    indicator: str,
    candles_two: List[Candle],
    indicator_cmp: str,
) -> bool:
    if len(candles) != len(candles_two):
        return False

    for i in range(len(candles) - 1, -1, -1):
        reading_one = reading_by_index(candles, indicator, i)
        reading_two = reading_by_index(candles_two, indicator_cmp, i)
        prev_one = reading_by_index(candles, indicator, i - 1)
        prev_two = reading_by_index(candles_two, indicator_cmp, i - 1)

        if None in [reading_one, reading_two, prev_one, prev_two]:
            continue
        elif (reading_one < reading_two and prev_one >= prev_two) or (
            reading_one > reading_two and prev_one <= prev_two
        ):
            return True

    return False


def crossover(
    candles: Indicator | Hexital | List[Candle],
    indicator: str,
    indicator_cmp: str,
    length: int = 1,
    index: int = -1,
) -> bool:
    """Crossover Analysis

    Determines whether the `indicator` reading has crossed over the `indicator_cmp` reading
    within a specified range of candles. A crossover occurs when `indicator` transitions from below
    to above `indicator_cmp` within the given range.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The primary indicator to evaluate.
        indicator_cmp (str): The secondary indicator to compare against.
        length (int, optional): The number of candles to include in the range. Defaults to 1.
            If `length` exceeds the total number of candles, all available candles are checked.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if `indicator` has crossed over `indicator_cmp` within the specified range;
        otherwise `False`.
    """
    candles_ = _retrieve_candles(candles, indicator, indicator_cmp)

    if not candles_:
        return False

    if len(candles_) == 1 and candles_[0]:
        candle_set = candles_[0]
        idx = absindex(index, len(candle_set)) + 1
        length = idx - (length + 1)

        return _crossover(candle_set[length:idx], indicator, candle_set[length:idx], indicator_cmp)
    elif len(candles_) == 2:
        candle_set = _timeframe_pair_candles(candles_)
        idx = absindex(index, len(candle_set[0])) + 1
        length = idx - (length + 1)

        return _crossover(
            candle_set[0][length:idx], indicator, candle_set[1][length:idx], indicator_cmp
        )

    return False


def _crossover(
    candles: List[Candle],
    indicator: str,
    candles_two: List[Candle],
    indicator_cmp: str,
) -> bool:
    if len(candles) != len(candles_two):
        return False

    for i in range(len(candles) - 1, -1, -1):
        reading_one = reading_by_index(candles, indicator, i)
        reading_two = reading_by_index(candles_two, indicator_cmp, i)
        prev_one = reading_by_index(candles, indicator, i - 1)
        prev_two = reading_by_index(candles_two, indicator_cmp, i - 1)

        if None in [reading_one, reading_two, prev_one, prev_two]:
            continue
        elif reading_one > reading_two and prev_one <= prev_two:
            return True

    return False


def crossunder(
    candles: Indicator | Hexital | List[Candle],
    indicator: str,
    indicator_cmp: str,
    length: int = 1,
    index: int = -1,
) -> bool:
    """Crossunder Analysis

    Determines whether the `indicator` reading has crossed under the `indicator_cmp` reading
    within a specified range of candles. A crossunder occurs when `indicator` transitions from above
    to below `indicator_cmp` within the given range.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The primary indicator to evaluate.
        indicator_cmp (str): The secondary indicator to compare against.
        length (int, optional): The number of candles to include in the range. Defaults to 1.
            If `length` exceeds the total number of candles, all available candles are checked.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if `indicator` has crossed under `indicator_cmp` within the specified range;
        otherwise `False`.
    """
    candles_ = _retrieve_candles(candles, indicator, indicator_cmp)

    if not candles_:
        return False

    if len(candles_) == 1 and candles_[0]:
        candle_set = candles_[0]
        idx = absindex(index, len(candle_set)) + 1
        length = idx - (length + 1)

        return _crossunder(
            candle_set[length:idx], indicator, candle_set[length:idx], indicator_cmp
        )
    elif len(candles_) == 2:
        candle_set = _timeframe_pair_candles(candles_)
        idx = absindex(index, len(candle_set[0])) + 1
        length = idx - (length + 1)

        return _crossunder(
            candle_set[0][length:idx], indicator, candle_set[1][length:idx], indicator_cmp
        )

    return False


def _crossunder(
    candles: List[Candle],
    indicator: str,
    candles_two: List[Candle],
    indicator_cmp: str,
) -> bool:
    if len(candles) != len(candles_two):
        return False

    for i in range(len(candles) - 1, -1, -1):
        reading_one = reading_by_index(candles, indicator, i)
        reading_two = reading_by_index(candles_two, indicator_cmp, i)
        prev_one = reading_by_index(candles, indicator, i - 1)
        prev_two = reading_by_index(candles_two, indicator_cmp, i - 1)

        if None in [reading_one, reading_two, prev_one, prev_two]:
            continue
        elif reading_one < reading_two and prev_one >= prev_two:
            return True

    return False


def flipped(
    candles: Indicator | Hexital | List[Candle], indicator: str, length: int = 1, index: int = -1
) -> bool:
    """Flipped Reading Analysis

    Determines whether the `indicator` has "flipped" its value, meaning the current reading is different
    from its previous reading, and within the last `length` candles, the indicator was above its previous reading.

    Args:
        candles (Indicator | Hexital | List[Candle]): The data source containing the indicators.
        indicator (str): The indicator to evaluate.
        length (int, optional): The number of candles to check for a flip. Defaults to 1.
            If `length` exceeds the total number of candles, all available candles are checked.
        index (int, optional): The index to start the evaluation. Defaults to -1 (latest candle).

    Returns:
        bool: `True` if the indicator has flipped (current value differs from previous); otherwise `False`.
    """
    candle_set = _retrieve_candles(candles, indicator)[0]
    if not candle_set:
        return False

    idx = absindex(index, len(candle_set))

    for idx in range(idx, idx - length, -1):
        if reading_by_index(candle_set, indicator, idx) != reading_by_index(
            candle_set, indicator, idx - 1
        ):
            return True

    return False



from dataclasses import dataclass, field
from typing import Optional

from hexital.core.indicator import Indicator, Managed
from hexital.indicators.atr import ATR
from hexital.indicators.rma import RMA


@dataclass(kw_only=True)
class ADX(Indicator):
    """Average Directional Index - ADX

    ADX is a trend strength in a series of prices of a financial instrument.

    Sources:
        https://en.wikipedia.org/wiki/Average_directional_movement_index

    Output type: `Dict["ADX": float, "DM_Plus": float, "DM_Neg": float]`

    Args:
        period (int): How many Periods to use. Defaults to 14
        period_signal (Optional[int]):  Average Directional Index period. Defaults same as period
        multiplier (Optional[float]): ADX smoothing multiplier. Defaults to 100.0
    """

    _name: str = field(init=False, default="ADX")

    period: int = 14
    period_signal: Optional[int] = None
    multiplier: float = 100.0

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}_{self.period_signal}"

    def _validate_fields(self):
        if self.period_signal is None:
            self.period_signal = self.period

    def _initialise(self):
        self.sub_atr = self.add_sub_indicator(
            ATR(
                period=self.period,
            )
        )

        self.data = self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

        self.sub_pos = self.data.add_sub_indicator(
            RMA(
                name=f"{self.name}_positive",
                period=self.period,
                source=f"{self.data.name}.positive",
            ),
            False,
        )
        self.sub_neg = self.data.add_sub_indicator(
            RMA(
                name=f"{self.name}_negative",
                period=self.period,
                source=f"{self.data.name}.negative",
            ),
            False,
        )
        self.dx = self.add_managed_indicator(
            "dx",
            RMA(
                name=f"{self.name}_dx",
                period=self.period_signal,
                source=f"{self.data.name}.dx",
            ),
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        adx_positive = None
        adx_negative = None

        if self.prev_exists("high"):
            up = self.candles[index].high - self.candles[index - 1].high
            down = self.candles[index - 1].low - self.candles[index].low
        else:
            up = 0
            down = 0

        dm_plus = ((up > down) & (up > 0)) * up
        dm_neg = ((down > up) & (down > 0)) * down

        self.data.set_reading({"positive": dm_plus, "negative": dm_neg})

        atr_ = self.sub_atr.reading()

        if self.sub_pos.exists() and atr_ is not None:
            mod = self.multiplier / atr_

            adx_positive = mod * self.sub_pos.reading()
            adx_negative = mod * self.sub_neg.reading()

            dx = self.multiplier * abs(adx_positive - adx_negative) / (adx_positive + adx_negative)

            self.data.set_reading({"positive": dm_plus, "negative": dm_neg, "dx": dx})
            self.dx.calculate_index(index)

        return {
            "ADX": self.dx.reading(),
            "DM_Plus": adx_positive,
            "DM_Neg": adx_negative,
        }



import inspect
from copy import copy
from typing import Callable, Optional

from hexital.core.indicator import Indicator


class Amorph(Indicator):
    """Amorph

    Flexible Skeleton Indicator that will use a method
    to generate readings on every Candle like indicators.

    The given Method is expected to have 'candles' and 'index' as named arguments, EG:

    Input type Example: [Doji][hexital.analysis.patterns.doji]

    Output type: Based on analysis method

    Args:
        analysis: Period to index back in
        args: All of the Arguments as keyword arguments as a dict of keyword arguments for called analysis
    """

    _analysis_method: Callable
    _analysis_kwargs: dict

    def __init__(self, analysis: Callable, args: Optional[dict] = None, **kwargs):
        self._analysis_method = analysis
        self._analysis_kwargs, kwargs = self._separate_indicator_attributes(kwargs)

        if isinstance(args, dict):
            self._analysis_kwargs.update(args)

        super().__init__(**kwargs)

    @property
    def settings(self) -> dict:
        """Returns a dict format of how this indicator can be generated"""
        output = {"analysis": self._analysis_method.__name__}

        for name, value in self.__dict__.items():
            if name == "candles":
                continue
            if name == "timeframe_fill" and self.timeframe is None:
                continue
            if not name.startswith("_") and value:
                output[name] = copy(value)

        return output

    @staticmethod
    def _separate_indicator_attributes(kwargs: dict) -> tuple[dict, dict]:
        indicator_attr = inspect.getmembers(Indicator)[1][1].keys()
        analysis_args = {}
        for argum in list(kwargs.keys()):
            if argum not in indicator_attr:
                analysis_args[argum] = kwargs.pop(argum)

        return analysis_args, kwargs

    def _generate_name(self) -> str:
        name = self._analysis_method.__name__
        period = self._analysis_kwargs.get("period")
        return f"{name}_{period}" if period else name

    def _calculate_reading(self, index: int) -> float | dict | None:
        return self._analysis_method(candles=self.candles, index=index, **self._analysis_kwargs)



from hexital.candlesticks import CANDLESTICK_MAP
from hexital.core.candlestick_type import CandlestickType
from hexital.exceptions import InvalidCandlestickType


def validate_candlesticktype(
    candlestick: CandlestickType | str,
) -> CandlestickType:
    if isinstance(candlestick, CandlestickType):
        return candlestick

    if not CANDLESTICK_MAP.get(candlestick):
        raise InvalidCandlestickType(f"Candlestick type {candlestick} is Invalid")

    requested_candlesticks = CANDLESTICK_MAP[candlestick]
    return requested_candlesticks()



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator, Managed


@dataclass(kw_only=True)
class CMO(Indicator):
    """Chande Momentum Oscillator - CMO

    The CMO indicator is created by calculating the difference between the
    sum of all recent higher closes and the sum of all recent lower closes
    and then dividing the result by the sum of all price movement over a given time period.
    The result is multiplied by 100 to give the -100 to +100 range.

    Sources:
        https://www.fidelity.com/learning-center/trading-investing/technical-analysis/technical-indicator-guide/cmo

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 14
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="CMO")
    period: int = 14
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.data = self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

    def _calculate_reading(self, index: int) -> float | dict | None:
        gains = None
        losses = None

        if self.prev_exists():
            change = self.prev_reading(self.source) - self.reading(self.source)

            change_gain = -1 * change if change < 0 else 0.0
            change_loss = change if change > 0 else 0.0

            gains = (
                (self.prev_reading(f"{self.data.name}.gain") * (self.period - 1)) + change_gain
            ) / self.period

            losses = (
                (self.data.prev_reading(f"{self.data.name}.loss") * (self.period - 1))
                + change_loss
            ) / self.period

        elif self.reading_period(self.period + 1, self.source):
            changes = [
                self.reading(self.source, i) - self.reading(self.source, i - 1)
                for i in range(index - (self.period - 1), index + 1)
            ]

            gains = sum(chng for chng in changes if chng > 0) / self.period
            losses = sum(abs(chng) for chng in changes if chng < 0) / self.period

        self.data.set_reading({"gain": gains, "loss": losses})

        if gains is not None and losses is not None:
            return ((gains - losses) / (gains + losses)) * 100



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class HLCA(Indicator):
    """High Low Close Average - HLCA

    Output type: `float`
    """

    _name: str = field(init=False, default="HLCA")

    def _generate_name(self) -> str:
        return f"{self._name}"

    def _calculate_reading(self, index: int) -> float | dict | None:
        return (self.candles[index].high + self.candles[index].low + self.candles[index].close) / 3



from dataclasses import dataclass, field

from hexital.analysis import movement
from hexital.core.indicator import Indicator, Managed
from hexital.indicators.sma import SMA


@dataclass(kw_only=True)
class STOCH(Indicator):
    """Stochastic - STOCH

    The Stochastic Oscillator (STOCH) was developed by George Lane in the 1950's.
    He believed this indicator was a good way to measure momentum because changes in
    momentum precede changes in price.

    It is a range-bound oscillator with two lines moving between 0 and 100.
    The first line (%K) displays the current close in relation to the period's
    high/low range. The second line (%D) is a Simple Moving Average of the %K line.
    The most common choices are a 14 period %K and a 3 period SMA for %D.

    %K = SMA(100 * (Current Close - Lowest Low) / (Highest High - Lowest Low), smoothK)
    %D = SMA(%K, periodD)

    Sources:
        https://www.tradingview.com/wiki/Stochastic_(STOCH)

    Output type: `Dict["stoch": float, "k": float, "d": float]`

    Args:
        period (int): How many Periods to use. Defaults to 14
        slow_period (int): How many Periods to use on smoothing d. Defaults to 3
        smoothing_k (int): How many Periods to use on smoothing K. Defaults to 3
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="STOCH")
    period: int = 14
    slow_period: int = 3
    smoothing_k: int = 3
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.data = self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))
        self.sub_k = self.data.add_sub_indicator(
            SMA(
                source=f"{self.name}_data.stoch",
                period=self.smoothing_k,
                name=f"{self.name}_k",
            ),
            False,
        )
        self.sub_d = self.add_managed_indicator(
            "STOCH_d",
            SMA(
                source=f"{self.name}_data.k",
                period=self.slow_period,
                name=f"{self.name}_d",
            ),
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        stoch = None
        k = None

        if self.reading_period(self.period, self.source):
            lowest = movement.lowest(self.candles, "low", self.period, index)
            highest = movement.highest(self.candles, "high", self.period, index)

            stoch = ((self.reading(self.source) - lowest) / (highest - lowest)) * 100

            self.data.set_reading({"stoch": stoch})
            k = self.sub_k.reading()

            self.data.set_reading({"stoch": stoch, "k": k})

            self.sub_d.calculate_index(index)

        return {"stoch": stoch, "k": k, "d": self.sub_d.reading()}



from datetime import datetime, timedelta
from enum import Enum
from typing import Optional

from hexital.exceptions import InvalidTimeFrame

VALID_TIMEFRAME_PREFIXES = ["S", "T", "H", "D"]


class TimeFrame(Enum):
    """Pre-defined TimeFrame values"""

    SECOND = "S1"
    SECOND5 = "S5"
    SECOND10 = "S10"
    SECOND15 = "S15"
    SECOND30 = "S30"
    MINUTE = "T1"
    MINUTE5 = "T5"
    MINUTE10 = "T10"
    MINUTE15 = "T15"
    MINUTE30 = "T30"
    MINUTE45 = "T45"
    HOUR = "H1"
    HOUR2 = "H2"
    HOUR3 = "H3"
    HOUR4 = "H4"
    DAY = "D1"
    WEEK = "D7"


def timeframe_validation(timeframe: Optional[str | TimeFrame | timedelta | int] = None) -> bool:
    if isinstance(timeframe, str):
        timeframe_ = timeframe.upper()
        if isinstance(timeframe_[0], str) and timeframe_[0] in VALID_TIMEFRAME_PREFIXES:
            if len(timeframe_) == 1:
                return True
            elif timeframe_[1].isdigit():
                return True
    elif isinstance(timeframe, (int, timedelta, TimeFrame)):
        return True

    return False


def convert_timeframe_to_timedelta(
    timeframe: Optional[str | TimeFrame | timedelta | int] = None,
) -> timedelta | None:
    if isinstance(timeframe, (str, TimeFrame)):
        return timeframe_to_timedelta(validate_timeframe(timeframe))
    elif isinstance(timeframe, int):
        return timedelta(seconds=timeframe)
    elif isinstance(timeframe, timedelta):
        return timeframe

    return None


def timeframe_to_timedelta(timeframe: str | TimeFrame) -> timedelta:
    # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

    timeframe_ = timeframe.value if isinstance(timeframe, TimeFrame) else timeframe.upper()

    if not timeframe_validation(timeframe_):
        raise InvalidTimeFrame(
            f"Invalid value: {timeframe_}, valid are: {VALID_TIMEFRAME_PREFIXES}, E.G 'T10' 10 minutes"
        )

    letter = timeframe_[0]
    time = 1 if len(timeframe_) == 1 else int(timeframe_[1:])

    if letter == "S":
        return timedelta(seconds=time)
    if letter == "T":
        return timedelta(minutes=time)
    if letter == "H":
        return timedelta(hours=time)
    if letter == "D":
        return timedelta(days=time)

    raise InvalidTimeFrame(f"Invalid value: {timeframe_}, somehow")


def timedelta_to_str(timeframe: timedelta) -> str:
    # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
    if not timeframe:
        return ""

    if timeframe < timedelta(seconds=60):
        return f"S{timeframe.seconds}"
    elif (
        timeframe < timedelta(minutes=60)
        or (timeframe < timedelta(hours=24) and timeframe.seconds / 60) % 60 != 0
    ):
        return f"T{int(timeframe.seconds / 60)}"
    elif (
        timeframe < timedelta(hours=24)
        or (timeframe >= timedelta(days=1) and timeframe.total_seconds() / 60 / 60) % 24 != 0
    ):
        return f"H{int(timeframe.total_seconds() / 60 / 60)}"
    elif timeframe >= timedelta(days=1):
        return f"D{int(timeframe.days)}"

    return ""


def validate_timeframe(timeframe: str | TimeFrame) -> str:
    if isinstance(timeframe, str):
        timeframe = timeframe.upper()
        if not isinstance(timeframe[0], str) or timeframe[0] not in VALID_TIMEFRAME_PREFIXES:
            raise InvalidTimeFrame(
                f"Invalid value: {timeframe}, valid are: {VALID_TIMEFRAME_PREFIXES}, E.G 'T10' 10 minutes"
            )
    elif isinstance(timeframe, TimeFrame):
        timeframe = timeframe.value

    return timeframe


def round_down_timestamp(timestamp: datetime, timeframe: timedelta) -> datetime:
    """Find and round down timestamp to the nearest matching timeframe. E.G timeframe of 5 minute
    E.G T5: 09:00:01 -> 9:00:00
    E.G T5: 09:01:20 -> 9:00:00
    E.G T5: 09:05:00 -> 9:05:00
    Note: This method also calls clean_timestamp, removing microseconds
    """
    timestamp = clean_timestamp(timestamp)
    if timeframe < timedelta(days=1):
        return datetime.fromtimestamp(
            timestamp.timestamp() // timeframe.total_seconds() * timeframe.total_seconds(),
            tz=timestamp.tzinfo,
        )
    elif timeframe < timedelta(days=7):
        return timestamp.replace(hour=0, minute=0, second=0)
    else:
        return timestamp.replace(day=0, hour=0, minute=0, second=0)


def within_timeframe(timestamp: datetime, within: datetime, timeframe: timedelta | None) -> bool:
    """Checks if timestamp is within other timestamp and timeframe period"""
    if not timeframe:
        return False
    return within - timeframe < timestamp <= within


def on_timeframe(timestamp: datetime, timeframe: timedelta) -> bool:
    """Checks if timestamp is on a timeframe value"""
    return timestamp.timestamp() % timeframe.total_seconds() == 0


def clean_timestamp(timestamp: datetime) -> datetime:
    """Removes Microseconds from the timestamp and returns it"""
    return timestamp.replace(microsecond=0)



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class OBV(Indicator):
    """On-Balance Volume - OBC

    On-balance volume (OBV) is a technical analysis indicator intended
    to relate price and volume in the stock market.
    OBV is based on a cumulative total volume.

    Sources:
       https://en.wikipedia.org/wiki/On-balance_volume

    Output type: `float`
    """

    _name: str = field(init=False, default="OBV")

    def _generate_name(self) -> str:
        return self._name

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists():
            if self.reading("close") == self.prev_reading("close"):
                return self.prev_reading()
            elif self.reading("close") > self.prev_reading("close"):
                return self.prev_reading() + self.reading("volume")

            return self.prev_reading() - self.reading("volume")

        return self.reading("volume")



import math
from dataclasses import dataclass, field

from hexital.core.indicator import Indicator, Managed


@dataclass(kw_only=True)
class JMA(Indicator):
    """Jurik Moving Average Average - JMA

    The JMA is an adaptive moving average that aims to reduce lag and improve responsiveness
    to price changes compared to traditional moving averages.
    By incorporating volatility and phase shift components, the JMA seeks to provide traders
    with a more accurate and timely representation of market trends.

    Sources:
        https://c.mql5.com/forextsd/forum/164/jurik_1.pdf

    Args:
        period (int): How many Periods to use. Defaults to 7
        source (str): Which input field to calculate the Indicator. Defaults to "close"
        phase (float): How heavy/light the average is [-100, 100]. Defaults to 0.0

    """

    _name: str = field(init=False, default="JMA")
    period: int = 7
    source: str = "close"
    phase: float = 0.0

    _phase_ratio: float = field(init=False, default=0)
    _beta: float = field(init=False, default=0)
    _length_1: float = field(init=False, default=0)
    _length_2: float = field(init=False, default=0)
    _power_1: float = field(init=False, default=0)
    _bet: float = field(init=False, default=0)

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}_{self.phase}"

    def _initialise(self):
        self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

    def _validate_fields(self):
        if self.phase > 100:
            self.phase = 100.0
        elif self.phase < -100:
            self.phase = -100.0
        self._phase_ratio = 1.5 + self.phase * 0.01

        self._beta = 0.45 * (self.period - 1) / (0.45 * (self.period - 1) + 2)

        length_ = 0.5 * (self.period - 1)
        self._length_1 = max((math.log(math.sqrt(length_)) / math.log(2.0)) + 2.0, 0)
        self._length_2 = self._length_1 * math.sqrt(length_)

        self._power_1 = max(self._length_1 - 2.0, 0.5)
        self._bet = self._length_2 / (self._length_2 + 1)

    def _calculate_reading(self, index: int) -> float | dict | None:
        price = self.reading(self.source)
        uband = self.prev_reading(f"{self.name}_data.uband", price)
        lband = self.prev_reading(f"{self.name}_data.lband", price)
        vsums = self.prev_reading(f"{self.name}_data.vsums", 0)
        ma_one = self.prev_reading(f"{self.name}_data.ma_one", price)
        ma_two = self.prev_reading(f"{self.name}_data.ma_two", 0)
        det_one = self.prev_reading(f"{self.name}_data.det_one", 0)
        det_two = self.prev_reading(f"{self.name}_data.det_two", 0)

        # Price Volatility
        del1 = price - uband
        del2 = price - lband
        volty = max(abs(del1), abs(del2)) if abs(del1) != abs(del2) else 0
        self.managed_indicators["data"].set_reading({"volty": volty})

        # Relative Price Volatility
        vsums = self.candles_average(10, f"{self.name}_data.volty")
        self.managed_indicators["data"].set_reading({"vsums": vsums, "volty": volty})

        avg_volty = self.candles_average(65, f"{self.name}_data.vsums")
        d_volty = 0 if avg_volty == 0 else volty / avg_volty
        r_volt = max(1.0, min(pow(self._length_1, 1 / self._power_1), d_volty))

        # Jurik Volatility Bands
        power_2 = pow(r_volt, self._power_1)
        kv = pow(self._bet, math.sqrt(power_2))
        uband = price if (del1 > 0) else price - (kv * del1)
        lband = price if (del2 < 0) else price - (kv * del2)

        power = pow(r_volt, self._power_1)
        alpha = pow(self._beta, power)

        # Stage One
        ma_one = ((1 - alpha) * price) + (alpha * ma_one)

        # Stage Two
        det_one = ((price - ma_one) * (1 - self._beta)) + (self._beta * det_one)
        ma_two = ma_one + self._phase_ratio * det_one

        # Stage Three
        det_two = ((ma_two - self.prev_reading(default=price)) * (1 - alpha) * (1 - alpha)) + (
            alpha * alpha * det_two
        )
        jma = self.prev_reading(default=price) + det_two

        self.managed_indicators["data"].set_reading(
            {
                "uband": uband,
                "lband": lband,
                "vsums": vsums,
                "volty": volty,
                "ma_one": ma_one,
                "ma_two": ma_two,
                "det_one": det_one,
                "det_two": det_two,
            }
        )

        return jma



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class TR(Indicator):
    """True Range - TR

    An method to expand a classical range (high minus low) to include
    possible gap scenarios.

    Sources:
        https://www.macroption.com/true-range/

    Output type: `float`
    """

    _name: str = field(init=False, default="TR")

    def _generate_name(self) -> str:
        return self._name

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists("close"):
            close = self.candles[index - 1].close
            high = self.candles[index].high
            low = self.candles[index].low
            return max(high - low, abs(high - close), abs(low - close))

        return None



class InvalidIndicator(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidAnalysis(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidTimeFrame(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidCandleOrder(Exception):
    def __init__(self, message):
        super().__init__(message)


class CandleAlreadyTagged(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidCandlestickType(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidConfiguration(Exception):
    def __init__(self, message):
        super().__init__(message)



from typing import List

from hexital.core.candle import Candle
from hexital.core.candlestick_type import CandlestickType


class HeikinAshi(CandlestickType):
    """Heikin-Ashi

    Heikin Ashi is a charting technique that can be used to predict future price movements.
    It is similar to traditional candlestick charts. However, unlike a regular candlestick
    chart, the Heikin Ashi chart tries to filter out some of the market noise by smoothing
    out strong price swings to better identify trend movements in the market.

    Sources:
        https://www.investopedia.com/trading/heikin-ashi-better-candlestick/
    """

    name: str = "Heikin-Ashi"
    acronym: str = "HA"

    def convert_candle(self, candle: Candle, candles: List[Candle], index: int):
        new_close = (candle.open + candle.high + candle.low + candle.close) / 4

        if index == 0:
            candle.open = (candle.open + candle.close) / 2
        else:
            candle.open = (candles[index - 1].open + candles[index - 1].close) / 2

        candle.high = max(candle.open, candle.high, new_close)
        candle.low = min(candle.open, candle.low, new_close)

        candle.close = new_close



from hexital.analysis import movement, patterns  # noqa F401
from hexital.core.candle import Candle  # noqa F401
from hexital.core.hexital import Hexital
from hexital.indicators import *  # noqa F401
from hexital.utils import TimeFrame  # noqa F401



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator
from hexital.indicators.tr import TR


@dataclass(kw_only=True)
class ATR(Indicator):
    """Average True Range - ATR

    Average True Range is used to measure volatility, especially volatility caused by
    gaps or limit moves.

    Sources:
        https://www.tradingview.com/wiki/Average_True_Range_(ATR)

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 14
    """

    _name: str = field(init=False, default="ATR")
    period: int = 14

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.sub_tr = self.add_sub_indicator(TR())

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists():
            return (self.prev_reading() * (self.period - 1) + self.sub_tr.reading()) / self.period

        if self.sub_tr.reading_period(self.period):
            return self.sub_tr.candles_average(self.period)

        return None



from dataclasses import dataclass, field
from datetime import timedelta
from typing import Optional

from hexital.core.indicator import Indicator, Managed
from hexital.exceptions import InvalidConfiguration
from hexital.utils.timeframe import (
    TimeFrame,
    convert_timeframe_to_timedelta,
    round_down_timestamp,
    timedelta_to_str,
    timeframe_validation,
)


@dataclass(kw_only=True)
class VWAP(Indicator):
    """Volume-Weighted Average Price - VWAP

    The volume-weighted average price is a technical analysis indicator
    used on intraday charts that resets at the start of every new trading session.

    Sources:
        https://www.investopedia.com/terms/v/vwap.asp

    Output type: `float`

    Args:
        anchor (Optional[str | TimeFrame | timedelta | int]): How to anchor VWAP, Depends on the index values, uses TimeFrame
    """

    _name: str = field(init=False, default="VWAP")
    anchor: Optional[str | TimeFrame | timedelta | int] = "D"

    def _generate_name(self) -> str:
        return f"{self._name}_{timedelta_to_str(self.anchor)}"

    def _validate_fields(self):
        if not timeframe_validation(self.anchor):
            raise InvalidConfiguration(f"Anchor is Invalid: {self.anchor}")

        self.anchor = convert_timeframe_to_timedelta(self.anchor)

    def _initialise(self):
        self.data = self.add_managed_indicator("VWAP_data", Managed(name=f"{self.name}_data"))

    def _calculate_reading(self, index: int) -> float | dict | None:
        candle = self.candles[index]

        current_anchor = round_down_timestamp(self.reading("timestamp"), self.anchor).timestamp()
        prev_anchor = self.prev_reading(f"{self.name}_data.active_anchor")
        typical_price = (candle.high + candle.low + candle.close) / 3.0

        if prev_anchor != current_anchor:
            pv = 0
            vol = 0
        else:
            pv = self.prev_reading(f"{self.name}_data.pv", 0.0)
            vol = self.prev_reading(f"{self.name}_data.vol", 0.0)

        pv = pv + (candle.volume * typical_price)
        vol = vol + candle.volume

        self.data.set_reading({"pv": pv, "vol": vol, "active_anchor": current_anchor})

        return pv / vol



from dataclasses import dataclass, field
from typing import Optional

from hexital.core.indicator import Indicator, Managed
from hexital.indicators.ema import EMA


@dataclass(kw_only=True)
class TSI(Indicator):
    """True Strength Index - TSI
    TSI attempts to show both trend direction and overbought/oversold conditions,
    using moving averages of the underlying momentum of a financial instrument.

    Sources:
        https://school.stockcharts.com/doku.php?id=technical_indicators:true_strength_index

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 25
        smooth_period (int): How much to smooth with EMA defaults: (period / 2) + (period % 2 > 0). Defaults to halve of period
        source (str): Which input field to calculate the Indicator. Defaults to "close"

    """

    _name: str = field(init=False, default="TSI")
    period: int = 25
    smooth_period: Optional[int] = None
    source: str = "close"

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}_{self.smooth_period}"

    def _validate_fields(self):
        if not self.smooth_period:
            self.smooth_period = int(int(self.period / 2) + (self.period % 2 > 0))

    def _initialise(self):
        self.data = self.add_managed_indicator("data", Managed(name=f"{self.name}_data"))

        self.data.add_sub_indicator(
            EMA(
                source=f"{self.name}_data.price",
                period=self.period,
                name=f"{self.name}_first",
            ),
            False,
        )

        self.sub_second = self.data.sub_indicators[f"{self.name}_first"].add_sub_indicator(
            EMA(
                source=f"{self.name}_first",
                period=self.smooth_period,
                name=f"{self.name}_second",
            ),
            False,
        )

        self.data.add_sub_indicator(
            EMA(
                source=f"{self.name}_data.abs_price",
                period=self.period,
                name=f"{self.name}_abs_first",
            ),
            False,
        )
        self.abs_second = self.data.sub_indicators[f"{self.name}_abs_first"].add_sub_indicator(
            EMA(
                source=f"{self.name}_abs_first",
                period=self.smooth_period,
                name=f"{self.name}_abs_second",
            ),
            False,
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        prev_reading = self.prev_reading(self.source)
        if prev_reading is None:
            return None

        source = self.reading(self.source)

        self.data.set_reading(
            {
                "price": source - prev_reading,
                "abs_price": abs(source - prev_reading),
            }
        )

        abs_second = self.abs_second.reading()
        if abs_second is not None:
            return 100 * (self.sub_second.reading() / abs_second)

        return None



from datetime import timedelta
from typing import List, Optional

from hexital.core.candle import Candle
from hexital.utils.indexing import absindex, valid_index
from hexital.utils.timeframe import round_down_timestamp


def reading_by_index(candles: List[Candle], name: str, index: int = -1) -> float | dict | None:
    """Simple method to get a reading from the given indicator from it's index"""
    if not valid_index(index, len(candles)):
        return None

    return reading_by_candle(candles[index], name)


def reading_by_candle(candle: Candle, name: str) -> float | dict | None:
    """Simple method to get a reading from the given indicator from a candle
    Uses '.' to find nested reading, E.G 'MACD_12_26_9.MACD"""

    if "." in name:
        main_name, nested_name = name.split(".")
        reading = _nested_indicator(candle, main_name, nested_name)
        return reading

    attr = getattr(candle, name, None)

    if attr is not None:
        return attr

    if name in candle.indicators:
        return candle.indicators[name]

    if name in candle.sub_indicators:
        return candle.sub_indicators[name]

    return None


def _nested_indicator(candle: Candle, name: str, nested_name: str) -> float | None:
    if name in candle.indicators:
        reading = candle.indicators[name]
        return reading.get(nested_name) if isinstance(reading, dict) else reading

    if name in candle.sub_indicators:
        reading = candle.sub_indicators[name]
        return reading.get(nested_name) if isinstance(reading, dict) else reading

    return None


def reading_count(candles: List[Candle], name: str, index: Optional[int] = None) -> int:
    """Returns how many instance of the given indicator exist"""
    index_ = absindex(index, len(candles))

    for count, idx in enumerate(range(index_, -1, -1)):
        if reading_by_candle(candles[idx], name) is None:
            return count

    return index_ + 1


def reading_period(
    candles: List[Candle], period: int, name: str, index: Optional[int] = None
) -> bool:
    """Will return True if the given indicator goes back as far as amount,
    It's true if exactly or more than. Includes index"""
    if not candles:
        return False

    period_ = period - 1
    index_ = absindex(index, len(candles))

    if index_ - period_ < 0:
        return False

    for point in [period_, period_ / 2, 0]:
        if reading_by_index(candles, name, index_ - int(point)) is None:
            return False
    return True


def candles_sum(
    candles: List[Candle],
    indicator: str,
    length: int,
    index: int = -1,
    include_latest: bool = True,
) -> float:
    """Sum of `indicator` for `length` bars back. If not enough Candles, sum's what's available"""
    return sum(get_readings_period(candles, indicator, length, index, include_latest))


def candles_average(
    candles: List[Candle],
    indicator: str,
    length: int,
    index: int = -1,
    include_latest: bool = True,
) -> float:
    """Averages period of `indicator` for `length` bars back.
    If not enough Candles, sum's what's available"""
    values = get_readings_period(candles, indicator, length, index, include_latest)
    return sum(values) / len(values) if values else 0


def get_readings_period(
    candles: List[Candle], indicator: str, length: int, index: int, include_latest: bool = False
) -> List[float | int]:
    """Goes through from index-length to index and returns a list of values, removes dict's and None values, validates index, if out of range set to max (-1)
    Returns from newest at the back (same order)"""
    index_ = absindex(index, len(candles))

    to_index = index_ + 1 if include_latest else index_

    start = to_index - length
    start = 0 if start < 0 else start

    candles_ = []

    for candle in candles[start:to_index]:
        reading = reading_by_candle(candle, indicator)
        if isinstance(reading, (float, int)):
            candles_.append(reading)

    return candles_


def get_candles_period(
    candles: List[Candle], length: int, index: int, include_latest: bool = False
) -> List[Candle]:
    """Goes through from index-length to index and returns a list of values
    Returns from newest at the back (same order)"""
    index_ = absindex(index, len(candles))

    to_index = index_ + 1 if include_latest else index_

    start = to_index - length
    start = 0 if start < 0 else start

    return candles[start:to_index]


def get_readings_timeframe(
    candles: List[Candle],
    indicator: str,
    timeframe: timedelta,
    index: int,
    include_latest: bool = False,
    rounded_timeframe: bool = False,
) -> List[float | int]:
    """Goes through from index and returns readings that are less or equal to timeframe to index candle
    removes dict's and None values
    Returns from newest at the back (same order)"""
    index_ = absindex(index, len(candles))

    to_index = index_ if include_latest else index_ - 1

    if rounded_timeframe:
        timestamp = round_down_timestamp(candles[index_].timestamp, timeframe)
        timestamp += timeframe
    else:
        timestamp = candles[index_].timestamp

    candles_ = []

    for candle in candles[to_index::-1]:
        if (timestamp - candle.timestamp) > timeframe:
            break
        reading = reading_by_candle(candle, indicator)
        if isinstance(reading, (float, int)):
            candles_.insert(0, reading)

    return candles_


def get_candles_timeframe(
    candles: List[Candle],
    timeframe: timedelta,
    index: int,
    include_latest: bool = False,
    rounded_timeframe: bool = False,
) -> List[Candle]:
    """Goes through from index and returns candles that are less or equal to timeframe to index candle
    Returns from newest at the back (same order)"""
    index_ = absindex(index, len(candles))

    to_index = index_ if include_latest else index_ - 1

    if rounded_timeframe:
        timestamp = round_down_timestamp(candles[index_].timestamp, timeframe)
        timestamp += timeframe
    else:
        timestamp = candles[index_].timestamp

    candles_ = []

    for candle in candles[to_index::-1]:
        if (timestamp - candle.timestamp) > timeframe:
            break
        candles_.insert(0, candle)

    return candles_



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator, Managed
from hexital.indicators.ema import EMA
from hexital.indicators.stdev import STDEV


@dataclass(kw_only=True)
class RVI(Indicator):
    """Relative Vigor Index - RVI

    The Relative Vigor Index, or RVI, is a popular member of the “Oscillator” family of technical indicators.
    although the creator of the Relative Vigor Index is unknown, its design is very similar to Stochastics
    except that the closing price is compared with the Open rather than the Low price for the period.

    Sources:
        https://www.thinkmarkets.com/en/learn-to-trade/indicators-and-patterns/indicators/relative-vigor-index-rvi-indicator/

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 14
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="RVI")
    period: int = 14
    source: str = "close"
    _scalar: float = field(init=False, default=100.0)

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _initialise(self):
        self.data = self.add_managed_indicator("data", Managed(name=f"{self._name}_data"))

        self.sub_stdev = self.add_sub_indicator(
            STDEV(
                source=self.source,
                period=self.period,
                name=f"{self._name}_stdev",
            ),
        )
        self.sub_pos = self.add_managed_indicator(
            "pos_ema",
            EMA(
                period=self.period,
                source=f"{self._name}_data.pos",
                name=f"{self._name}_pos_ema",
            ),
        )
        self.sub_neg = self.add_managed_indicator(
            "neg_ema",
            EMA(
                period=self.period,
                source=f"{self._name}_data.neg",
                name=f"{self._name}_neg_ema",
            ),
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        stdev_reading = self.sub_stdev.reading()

        if stdev_reading is not None:
            prev_reading = self.prev_reading(self.source)

            if prev_reading is None:
                return None

            cur_reading = self.reading(self.source)

            pos = 0 if cur_reading <= prev_reading else 1
            neg = 0 if cur_reading >= prev_reading else 1

            pos_stdev = pos * stdev_reading
            neg_stdev = neg * stdev_reading

            self.data.set_reading({"pos": pos_stdev, "neg": neg_stdev})
            self.sub_pos.calculate_index(index)
            self.sub_neg.calculate_index(index)

        positive_smoothed = self.sub_pos.reading()
        negative_smoothed = self.sub_neg.reading()

        if self.prev_exists() or positive_smoothed is not None:
            return (self._scalar * positive_smoothed) / (positive_smoothed + negative_smoothed)

        return None



from .movement import (  # noqa F401
    cross,
    crossover,
    crossunder,
    falling,
    highest,
    highestbar,
    lowest,
    lowestbar,
    mean_falling,
    mean_rising,
    negative,
    positive,
    rising,
    value_range,
)
from .patterns import doji, dojistar, hammer, inverted_hammer  # noqa F401



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator


@dataclass(kw_only=True)
class RMA(Indicator):
    """wildeR's Moving Average - RMA

    Wilder's Moving Average places more emphasis on recent price movements than other moving averages.
    This makes it a more responsive tool for short-term traders who need to adapt quickly to
    changing market conditions.

    Sources:
        https://tlc.thinkorswim.com/center/reference/Tech-Indicators/studies-library/V-Z/WildersSmoothing
        https://www.incrediblecharts.com/indicators/wilder_moving_average.php

    Output type: `float`

    Args:
        period (int): How many Periods to use. Defaults to 10
        source (str): Which input field to calculate the Indicator. Defaults to "close"
    """

    _name: str = field(init=False, default="RMA")
    period: int = 10
    source: str = "close"
    _alpha: float = field(init=False, default=0)

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}"

    def _validate_fields(self):
        self._alpha = float(1.0 / self.period)

    def _calculate_reading(self, index: int) -> float | dict | None:
        if self.prev_exists():
            return float(
                (self._alpha * self.reading(self.source))
                + ((1.0 - self._alpha) * self.prev_reading())
            )

        if self.reading_period(self.period, self.source):
            period_to = index - self.period

            # numpy ewm adjusted calc
            values = sum(
                ((1 - self._alpha) ** py) * self.reading(self.source, i)
                for py, i in enumerate(range(index, period_to, -1))
            )

            divide_by = sum(
                (1 - self._alpha) ** i for py, i in enumerate(range(index, period_to, -1))
            )

            return values / divide_by

        return None



from copy import copy
from datetime import timedelta
from importlib import import_module
from typing import Dict, List, Optional, Set

from hexital.core.candle import Candle
from hexital.core.candle_manager import DEFAULT_CANDLES, CandleManager
from hexital.core.candlestick_type import CandlestickType
from hexital.core.indicator import Indicator
from hexital.exceptions import (
    InvalidAnalysis,
    InvalidIndicator,
)
from hexital.indicators.amorph import Amorph
from hexital.utils.candles import reading_by_candle, reading_by_index
from hexital.utils.candlesticks import validate_candlesticktype
from hexital.utils.timeframe import (
    TimeFrame,
    convert_timeframe_to_timedelta,
    timedelta_to_str,
    timeframe_validation,
)


class Hexital:
    name: str
    _candles: Dict[str, CandleManager]
    _indicators: Dict[str, Indicator]
    description: Optional[str] = None
    _timeframe: Optional[timedelta] = None
    _timeframe_name: str = DEFAULT_CANDLES
    timeframe_fill: bool = False
    candle_life: Optional[timedelta] = None
    candlestick: Optional[CandlestickType] = None

    def __init__(
        self,
        name: str,
        candles: List[Candle],
        indicators: Optional[List[dict | Indicator]] = None,
        description: Optional[str] = None,
        timeframe: Optional[str | TimeFrame | timedelta | int] = None,
        timeframe_fill: bool = False,
        candle_life: Optional[timedelta] = None,
        candlestick: Optional[CandlestickType | str] = None,
    ):
        self.name = name
        self.description = description

        self._timeframe = convert_timeframe_to_timedelta(timeframe)
        self.timeframe_fill = timeframe_fill
        self.candle_life = candle_life

        if candlestick:
            self.candlestick = validate_candlesticktype(candlestick)

        if self._timeframe:
            self._timeframe_name = timedelta_to_str(self._timeframe)

        self._candles = {
            self._timeframe_name: CandleManager(
                candles if isinstance(candles, list) else [],
                candle_life=self.candle_life,
                timeframe=self._timeframe,
                timeframe_fill=self.timeframe_fill,
                candlestick=self.candlestick,
            )
        }

        self._indicators = self._validate_indicators(indicators) if indicators else {}

    @property
    def timeframe(self) -> str | None:
        return self._timeframe_name if self._timeframe else None

    @property
    def timeframes(self) -> Set[str]:
        return {manager.name for manager in self._candles.values()}

    @property
    def indicators(self) -> Dict[str, Indicator]:
        """Simply get's a list of all the Indicators within Hexital strategy"""
        return self._indicators

    def indicator(self, name: str) -> Indicator:
        """Searches hexital's indicator's and Returns the Indicator object itself."""
        return self._indicators[name]

    def has_reading(self, name: str) -> bool:
        """Checks if the given Indicator has a valid reading in latest Candle"""
        value = self.reading(name)
        if isinstance(value, dict):
            return any(v is not None for v in value.values())
        return value is not None

    def candles(self, name: Optional[str | TimeFrame | timedelta | int] = None) -> List[Candle]:
        """Get a set of candles by using either a Timeframe or Indicator name"""
        name_ = name if name else self._timeframe_name
        timeframe_name = self._parse_timeframe(name)

        name_ = timeframe_name if timeframe_name else name_

        if isinstance(name_, str) and self._candles.get(name_, False):
            return self._candles[name_].candles
        elif isinstance(name_, str):
            for manager in self._candles.values():
                if manager.find_indicator(name_):
                    return manager.candles

        return []

    def get_candles(self) -> Dict[str, List[Candle]]:
        return {name: manager.candles for name, manager in self._candles.items()}

    @property
    def indicator_settings(self) -> List[dict]:
        """Simply get's a list of all the Indicators within Hexital strategy"""
        return [indicator.settings for indicator in self._indicators.values()]

    def reading(self, name: str, index: int = -1) -> float | dict | None:
        """Attempts to retrieve a reading with a given Indicator name.
        `name` can use '.' to find nested reading, E.G `MACD_12_26_9.MACD`
        """
        reading = reading_by_index(self._candles[self._timeframe_name].candles, name, index=index)

        if reading is not None:
            return reading

        for candle_manager in self._candles.values():
            reading = reading_by_index(candle_manager.candles, name, index=index)
            if reading is not None:
                return reading

        return None

    def prev_reading(self, name: str) -> float | dict | None:
        return self.reading(name, index=-2)

    def readings(self) -> Dict[str, List[float | dict | None]]:
        """Returns a Dictionary of all the Indicators and there results in a list format."""

        return {name: indicator.as_list() for name, indicator in self._indicators.items()}

    def reading_as_list(self, name: str) -> List[float | dict | None]:
        """Find given indicator and returns the readings as a list
        Full Name of the indicator E.G `EMA_12` OR `MACD_12_26_9.MACD`"""
        primary_name = name.split(".")[0]
        if self._indicators.get(primary_name):
            return self._indicators[primary_name].as_list(name)
        return []

    def add_indicator(
        self, indicator: Indicator | List[Indicator | Dict[str, str]] | Dict[str, str]
    ):
        """Add's a new indicator to `Hexital` strategy.
        This accept either `Indicator` datatypes or dict string versions to be packed.
        `add_indicator(SMA(period=10))` or `add_indicator({"indicator": "SMA", "period": 10})`
        Does not automatically calculates readings."""
        indicators = indicator if isinstance(indicator, list) else [indicator]

        for valid_indicator in self._validate_indicators(indicators).values():
            self._indicators[valid_indicator.name] = valid_indicator

    def remove_indicator(self, name: str):
        """Removes an indicator from running within hexital"""
        self.purge(name)
        self._indicators.pop(name, None)

    def append(
        self,
        candles: Candle | List[Candle] | dict | List[dict] | list | List[list],
        timeframe: Optional[str | TimeFrame | timedelta | int] = None,
    ):
        timeframe_name = self._parse_timeframe(timeframe)

        if timeframe_name and self._candles.get(timeframe_name):
            self._candles[timeframe_name].append(candles)
        else:
            for candle_manager in self._candles.values():
                candle_manager.append(candles)

        self.calculate()

    def calculate(self, name: Optional[str] = None):
        """Calculates all the missing indicator readings."""
        for indicator_name, indicator in self._indicators.items():
            if name is None or indicator_name == name:
                indicator.calculate()

    def calculate_index(self, name: Optional[str] = None, index: int = -1):
        """Calculate specific index for all or specific indicator readings."""
        for indicator_name, indicator in self._indicators.items():
            if name is None or indicator_name == name:
                indicator.calculate_index(index)

    def recalculate(self, name: Optional[str] = None):
        """Purge's all indicator reading's and re-calculates them all,
        ideal for changing an indicator parameters midway."""
        self.purge(name)
        self.calculate(name)

    def purge(self, name: Optional[str] = None):
        """Takes Indicator name and removes all readings for said indicator.
        Indicator name must be exact"""
        for indicator_name, indicator in self._indicators.items():
            if name is None or (name and name in indicator_name):
                indicator.purge()

    def _parse_timeframe(
        self, timeframe: Optional[str | TimeFrame | timedelta | int]
    ) -> str | None:
        if not timeframe:
            return None

        if timeframe == self._timeframe_name:
            return self._timeframe_name

        if not timeframe_validation(timeframe):
            return None

        name = convert_timeframe_to_timedelta(timeframe)

        return None if not name else timedelta_to_str(name)

    def _validate_indicators(self, indicators: List[dict | Indicator]) -> Dict[str, Indicator]:
        if not indicators:
            return {}

        valid_indicators = {}

        for indicator in indicators:
            if isinstance(indicator, Indicator):
                valid_indicators[indicator.name] = indicator
                continue

            if not isinstance(indicator, dict):
                raise InvalidIndicator(
                    f"Indicator type invalid 'indicator' must be a dict or Indicator type: {indicator}"
                )

            new_indicator = self._build_indicator(indicator)
            valid_indicators[new_indicator.name] = new_indicator

        for indicator in valid_indicators.values():
            if not indicator.timeframe:
                indicator.candle_manager = self._candles[self._timeframe_name]
            elif indicator.timeframe and indicator.timeframe in self._candles:
                indicator.candle_manager = self._candles[indicator.timeframe]
            else:
                manager = CandleManager(
                    [],
                    candle_life=self.candle_life,
                    timeframe=indicator._timeframe if indicator._timeframe else self._timeframe,
                    timeframe_fill=self.timeframe_fill,
                    candlestick=self.candlestick,
                )
                manager.append(self._candles[self._timeframe_name].candles)
                self._candles[manager.name] = manager
                indicator.candle_manager = self._candles[manager.name]

        return valid_indicators

    def _build_indicator(self, raw_indicator: dict) -> Indicator:
        indicator = copy(raw_indicator)

        if indicator.get("indicator"):
            indicator_name = indicator.pop("indicator")
            indicator_class = getattr(import_module("hexital.indicators"), indicator_name, None)

            if indicator_class:
                return indicator_class(**indicator)
            else:
                raise InvalidIndicator(
                    f"Indicator {indicator_name} does not exist. [{raw_indicator}]"
                )

        elif indicator.get("analysis") and isinstance(indicator.get("analysis"), str):
            analysis_name = indicator.pop("analysis")
            analysis_class = getattr(import_module("hexital.analysis"), analysis_name, None)

            if not analysis_class:
                raise InvalidAnalysis(
                    f"analysis {analysis_name} does not exist in patterns or movements. [{raw_indicator}]"
                )

            return Amorph(analysis=analysis_class, **indicator)

        elif indicator.get("analysis") and callable(indicator.get("analysis")):
            method_name = indicator.pop("analysis")
            return Amorph(analysis=method_name, **indicator)
        else:
            raise InvalidAnalysis(
                f"Dict Indicator missing 'indicator' or 'analysis' name, not: {raw_indicator}"
            )

    def find_candles(
        self, indicator: str, indicator_cmp: Optional[str] = None
    ) -> List[List[Candle]]:
        reverted = False

        if indicator and not indicator_cmp:
            return [self.candles(indicator), []]

        if indicator_cmp and indicator in ["open", "high", "low", "close", "volume"]:
            indicator, indicator_cmp = indicator_cmp, indicator
            reverted = True

        candles = self.candles(indicator)

        if candles and indicator_cmp and reading_by_candle(candles[-1], indicator_cmp) is not None:
            return [candles, candles]
        elif candles and indicator_cmp and reading_by_candle(candles[-1], indicator_cmp) is None:
            if reverted:
                return [self.candles(indicator_cmp), candles]
            else:
                return [candles, self.candles(indicator_cmp)]

        return [[]]



from dataclasses import dataclass, field

from hexital.core.indicator import Indicator
from hexital.indicators import ATR, EMA


@dataclass(kw_only=True)
class KC(Indicator):
    """Keltner Channel - KC

    Keltner channel is a technical analysis indicator showing a central moving
    average line plus channel lines at a distance above and below.
    A popular volatility indicator similar to Bollinger Bands and Donchian Channels.

    Sources:
        https://www.investopedia.com/terms/k/keltnerchannel.asp

    Output type: `Dict["lower": float, "band": float, "upper": float]`

    Args:
        period (int): How many Periods to use. Defaults to 20
        source (str): Which input field to calculate the Indicator. Defaults to "close"
        multiplier (float): A positive float to multiply the bands. Defaults to 2.0
    """

    _name: str = field(init=False, default="KC")
    period: int = 20
    source: str = "close"
    multiplier: float = 2.0

    def _generate_name(self) -> str:
        return f"{self._name}_{self.period}_{self.multiplier}"

    def _initialise(self):
        self.sub_atr = self.add_sub_indicator(ATR(period=self.period))

        self.sub_ema = self.add_sub_indicator(
            EMA(
                source=self.source,
                period=self.period,
                name=f"{self.name}_EMA",
            )
        )

    def _calculate_reading(self, index: int) -> float | dict | None:
        atr_ = self.sub_atr.reading()
        ema_ = self.sub_ema.reading()

        if atr_ is None:
            return {"lower": None, "band": ema_, "upper": None}

        lower = ema_ - (self.multiplier * atr_)
        upper = ema_ + (self.multiplier * atr_)

        return {
            "lower": lower,
            "band": ema_,
            "upper": upper,
        }
